# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_config import cfg
import oslo_messaging as messaging
from oslo_messaging.rpc import dispatcher

from ironic_inspector.conductor import manager

CONF = cfg.CONF
TRANSPORT = None


def init():
    global TRANSPORT
    TRANSPORT = messaging.get_rpc_transport(CONF)


def get_client(topic=None):
    """Get a RPC client instance.

    :param topic: The topic of the message will be delivered to. This argument
                  is ignored if CONF.standalone is True.
    """
    assert TRANSPORT is not None
    if CONF.standalone:
        target = messaging.Target(topic=manager.MANAGER_TOPIC,
                                  server=CONF.host,
                                  version='1.3')
    else:
        target = messaging.Target(topic=topic, version='1.3')
    return messaging.get_rpc_client(TRANSPORT, target)


def get_server(endpoints):
    """Get a RPC server instance."""

    assert TRANSPORT is not None
    target = messaging.Target(topic=manager.MANAGER_TOPIC, server=CONF.host,
                              version='1.3')
    return messaging.get_rpc_server(
        TRANSPORT, target, endpoints, executor='eventlet',
        access_policy=dispatcher.DefaultRPCAccessPolicy)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Port Physical Network Hook"""

import ipaddress

from oslo_config import cfg

from ironic_inspector.plugins import base_physnet

CONF = cfg.CONF


class PhysnetCidrMapHook(base_physnet.BasePhysnetHook):
    """Process port physical network

    Set the physical_network field of baremetal ports based on a cidr to
    physical network mapping in the configuration.
    """

    def get_physnet(self, port, iface_name, introspection_data):
        """Return a physical network to apply to a port.

        :param port: The ironic port to patch.
        :param iface_name: Name of the interface.
        :param introspection_data: Introspection data.
        :returns: The physical network to set, or None.
        """

        def get_iface_ips(iface):
            ips = []
            for addr_version in ['ipv4_address', 'ipv6_address']:
                try:
                    ips.append(ipaddress.ip_address(iface.get(addr_version)))
                except ValueError:
                    pass

            return ips

        # Convert list config to a dict with ip_networks as keys
        cidr_map = {
            ipaddress.ip_network(x.rsplit(':', 1)[0]): x.rsplit(':', 1)[1]
            for x in CONF.port_physnet.cidr_map}

        iface = [i for i in introspection_data['inventory']['interfaces']
                 if i['name'] == iface_name][0]
        ips = get_iface_ips(iface)

        for ip in ips:
            try:
                return [cidr_map[cidr] for cidr in cidr_map if ip in cidr][0]
            except IndexError:
                # No mapping found for any of the ip addresses
                return None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_config import cfg

from ironic_inspector.common import coordination
from ironic_inspector.common import locking
from ironic_inspector.test import base as test_base

CONF = cfg.CONF


@mock.patch.object(locking, 'InternalLock', autospec=True)
@mock.patch.object(locking, 'ToozLock', autospec=True)
class TestGetLock(test_base.NodeTest):
    def test_get_lock_internal(self, mock_tooz, mock_internal):
        locking.get_lock(self.node.uuid)

        mock_internal.assert_called_once_with(self.node.uuid)
        mock_tooz.assert_not_called()

    @mock.patch.object(coordination, 'get_coordinator', autospec=True)
    def test_get_lock_tooz(self, mock_get_coord, mock_tooz, mock_internal):
        CONF.set_override('standalone', False)
        mock_lock = mock.Mock()
        coordinator = mock.Mock()
        coordinator.get_lock.return_value = mock_lock
        mock_get_coord.return_value = coordinator

        locking.get_lock(self.node.uuid)

        mock_tooz.assert_called_once_with(mock_lock)
        mock_internal.assert_not_called()


class TestInternalLock(test_base.NodeTest):
    def setUp(self):
        super(TestInternalLock, self).setUp()
        self.mock_lock = mock.MagicMock()
        self.mock_lock.acquire.return_value = True

    @mock.patch.object(locking, 'lockutils', autospec=True)
    def test_init_lock(self, mock_lockutils):
        locking.InternalLock(self.node.uuid)
        mock_lockutils.internal_lock.assert_called_with(
            'node-%s' % self.node.uuid, mock.ANY)

    def test_acquire(self):
        lock = locking.InternalLock(self.node.uuid)
        lock._lock = self.mock_lock
        lock.acquire()
        self.mock_lock.acquire.assert_called_once_with(blocking=True)

    def test_release(self):
        lock = locking.ToozLock(self.mock_lock)
        self.mock_lock._locked = True

        lock.release()

        self.mock_lock.release.assert_called_once_with()

    def test_context(self):
        lock = locking.InternalLock(self.node.uuid)
        lock._lock = self.mock_lock

        with lock:
            self.mock_lock.acquire.assert_called_once_with()

        self.mock_lock.release.assert_called_once_with()


class TestToozLock(test_base.NodeTest):
    def setUp(self):
        super(TestToozLock, self).setUp()
        self.mock_lock = mock.MagicMock()
        self.mock_lock.acquire.return_value = True
        self.mock_lock.acquired = False

    def test_acquire(self):
        lock = locking.ToozLock(self.mock_lock)
        lock.acquire()
        self.mock_lock.acquire.assert_called_once_with(blocking=True)

    def test_release(self):
        self.mock_lock.acquired = True
        lock = locking.ToozLock(self.mock_lock)

        lock.release()

        self.mock_lock.release.assert_called_once_with()

    def test_context(self):
        lock = locking.ToozLock(self.mock_lock)

        with lock:
            self.mock_lock.acquire.assert_called_once_with()

        self.mock_lock.release.assert_called_once_with()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_introspection_data_table

Revision ID: bf8dec16023c
Revises: 2970d2d44edc
Create Date: 2018-07-19 18:51:38.124614

"""

from alembic import op
from oslo_db.sqlalchemy import types as db_types
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'bf8dec16023c'
down_revision = '2970d2d44edc'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'introspection_data',
        sa.Column('uuid', sa.String(36), sa.ForeignKey('nodes.uuid'),
                  primary_key=True),
        sa.Column('processed', sa.Boolean, default=False, primary_key=True),
        sa.Column('data', db_types.JsonEncodedDict(mysql_as_long=True).impl,
                  nullable=True),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Added 'scope' column to 'Rules' table

Revision ID: b55109d5063a
Revises: bf8dec16023c
Create Date: 2019-12-11 14:15:57.510289

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'b55109d5063a'
down_revision = 'bf8dec16023c'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('rules', sa.Column('scope', sa.String(255),
                                     nullable=True, default=None))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import os
import tempfile
import unittest
from unittest import mock

import fixtures
import oslo_messaging as messaging
from oslo_utils import uuidutils

from ironic_inspector.common import coordination
from ironic_inspector.common import ironic as ir_utils
from ironic_inspector.common import rpc
from ironic_inspector.common import swift
import ironic_inspector.conf
from ironic_inspector.conf import opts as conf_opts
from ironic_inspector import introspection_state as istate
from ironic_inspector import main
from ironic_inspector import node_cache
from ironic_inspector.plugins import base as plugins_base
from ironic_inspector.plugins import example as example_plugin
from ironic_inspector.plugins import introspection_data as intros_data_plugin
from ironic_inspector import rules
from ironic_inspector.test import base as test_base
from ironic_inspector import utils

CONF = ironic_inspector.conf.CONF


def _get_error(res):
    return json.loads(res.data.decode('utf-8'))['error']['message']


class BaseAPITest(test_base.BaseTest):

    def init_app(self):
        CONF.set_override('auth_strategy', 'noauth')
        main._app.testing = True
        self.app = main.get_app().test_client()
        self.headers = {}

    def setUp(self):
        super(BaseAPITest, self).setUp()
        self.init_app()
        self.uuid = uuidutils.generate_uuid()
        self.rpc_get_client_mock = self.useFixture(
            fixtures.MockPatchObject(rpc, 'get_client', autospec=True)).mock
        self.client_mock = mock.MagicMock(spec=messaging.RPCClient)
        self.rpc_get_client_mock.return_value = self.client_mock


class TestApiIntrospect(BaseAPITest):

    def test_root_endpoint_with_prefix(self):
        res = self.app.get(
            '/', headers={'X-Forwarded-Proto': 'https',
                          'X-Forwarded-Host': 'myhost',
                          'X-Forwarded-Prefix': '/ironic_inspector'})

        self.assertEqual(200, res.status_code)
        self.assertEqual('application/json',
                         res.headers['content-type'])
        cur_version = res.json['versions'][0]
        hrefs = [link['href'] for link in cur_version['links']]
        for href in hrefs:
            self.assertTrue(href.startswith('https://myhost/ironic_inspector'))

    def test_introspect(self):

        res = self.app.post('/v1/introspection/%s' % self.uuid,
                            headers=self.headers)

        self.assertEqual(202, res.status_code)
        self.client_mock.call.assert_called_once_with({}, 'do_introspection',
                                                      node_id=self.uuid,
                                                      manage_boot=True,
                                                      token=None)
        self.assertEqual('application/json',
                         res.headers['content-type'])
        self.assertEqual(b'{}\n', res.data)

    def test_intospect_failed(self):
        self.client_mock.call.side_effect = utils.Error("boom")
        res = self.app.post('/v1/introspection/%s' % self.uuid,
                            headers=self.headers)

        self.assertEqual(400, res.status_code)
        self.assertEqual(
            'boom',
            json.loads(res.data.decode('utf-8'))['error']['message'])
        self.client_mock.call.assert_called_once_with({}, 'do_introspection',
                                                      node_id=self.uuid,
                                                      manage_boot=True,
                                                      token=None)

    def test_introspect_no_manage_boot(self):
        res = self.app.post('/v1/introspection/%s?manage_boot=0' % self.uuid,
                            headers=self.headers)
        self.assertEqual(202, res.status_code)
        self.client_mock.call.assert_called_once_with({}, 'do_introspection',
                                                      node_id=self.uuid,
                                                      manage_boot=False,
                                                      token=None)
        self.assertEqual('application/json',
                         res.headers['content-type'])
        self.assertEqual(b'{}\n', res.data)

    def test_introspect_can_manage_boot_false(self):
        CONF.set_override('can_manage_boot', False)
        res = self.app.post('/v1/introspection/%s?manage_boot=0' % self.uuid,
                            headers=self.headers)
        self.assertEqual(202, res.status_code)
        self.client_mock.call.assert_called_once_with({}, 'do_introspection',
                                                      node_id=self.uuid,
                                                      manage_boot=False,
                                                      token=None)

    def test_introspect_can_manage_boot_false_failed(self):
        CONF.set_override('can_manage_boot', False)
        res = self.app.post('/v1/introspection/%s' % self.uuid,
                            headers=self.headers)
        self.assertEqual(400, res.status_code)
        self.assertFalse(self.client_mock.call.called)

    def test_introspect_wrong_manage_boot(self):
        res = self.app.post('/v1/introspection/%s?manage_boot=foo' % self.uuid,
                            headers=self.headers)
        self.assertEqual(400, res.status_code)
        self.assertFalse(self.client_mock.call.called)
        self.assertEqual(
            'Invalid boolean value for manage_boot: foo',
            json.loads(res.data.decode('utf-8'))['error']['message'])

    @mock.patch.object(utils, 'check_auth', autospec=True)
    def test_introspect_failed_authentication(self, auth_mock):
        CONF.set_override('auth_strategy', 'keystone')
        auth_mock.side_effect = utils.Error('Boom', code=403)
        res = self.app.post('/v1/introspection/%s' % self.uuid,
                            headers={'X-Auth-Token': 'token'})
        self.assertEqual(403, res.status_code)
        self.assertFalse(self.client_mock.call.called)


class TestBasicAuthApiIntrospect(TestApiIntrospect):

    def init_app(self):
        CONF.set_override('auth_strategy', 'http_basic')
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
            f.write('myName:$2y$05$lE3eGtyj41jZwrzS87KTqe6.'
                    'JETVCWBkc32C63UP2aYrGoYOEpbJm\n\n\n')
            self.addCleanup(os.remove, f.name)
            CONF.set_override('http_basic_auth_user_file', f.name)
        main._app.config['TESTING'] = True
        self.app = main.get_app().test_client()

        # base64 encode myName:myPassword
        self.headers = {'Authorization': 'Basic bXlOYW1lOm15UGFzc3dvcmQ='}

    def test_introspect_failed_authentication(self):
        # base64 encode myName:yourPassword
        self.headers = {'Authorization': 'Basic bXlOYW1lOnlvdXJQYXNzd29yZA=='}

        res = self.app.post('/v1/introspection/%s' % self.uuid,
                            headers=self.headers)
        self.assertEqual(401, res.status_code)

    def test_unauthenticated_public_api(self):
        res = self.app.get('/')
        self.assertEqual(200, res.status_code)
        res = self.app.get('/v1')
        self.assertEqual(200, res.status_code)
        res = self.app.get('/v1/introspection')
        self.assertEqual(401, res.status_code)


class TestApiContinue(BaseAPITest):
    def test_continue(self):
        # should be ignored
        CONF.set_override('auth_strategy', 'keystone')
        self.client_mock.call.return_value = {'result': 42}
        res = self.app.post('/v1/continue', data='{"foo": "bar"}')
        self.assertEqual(200, res.status_code)
        self.client_mock.call.assert_called_once_with({}, 'do_continue',
                                                      data={"foo": "bar"})
        self.assertEqual({"result": 42}, json.loads(res.data.decode()))

    def test_continue_failed(self):
        self.client_mock.call.side_effect = utils.Error("boom")
        res = self.app.post('/v1/continue', data='{"foo": "bar"}')
        self.assertEqual(400, res.status_code)
        self.client_mock.call.assert_called_once_with({}, 'do_continue',
                                                      data={"foo": "bar"})
        self.assertEqual('boom', _get_error(res))

    def test_continue_wrong_type(self):
        res = self.app.post('/v1/continue', data='42')
        self.assertEqual(400, res.status_code)
        self.assertEqual('Invalid data: expected a JSON object, got int',
                         _get_error(res))
        self.client_mock.call.assert_not_called()


class TestApiAbort(BaseAPITest):
    def test_ok(self):

        res = self.app.post('/v1/introspection/%s/abort' % self.uuid,
                            headers={'X-Auth-Token': 'token'})

        self.client_mock.call.assert_called_once_with({}, 'do_abort',
                                                      node_id=self.uuid,
                                                      token='token')
        self.assertEqual(202, res.status_code)
        self.assertEqual(b'{}\n', res.data)
        self.assertEqual('application/json',
                         res.headers['content-type'])

    def test_no_authentication(self):

        res = self.app.post('/v1/introspection/%s/abort' % self.uuid)

        self.client_mock.call.assert_called_once_with({}, 'do_abort',
                                                      node_id=self.uuid,
                                                      token=None)
        self.assertEqual(202, res.status_code)
        self.assertEqual(b'{}\n', res.data)
        self.assertEqual('application/json',
                         res.headers['content-type'])

    def test_node_not_found(self):
        exc = utils.Error("Not Found.", code=404)
        self.client_mock.call.side_effect = exc

        res = self.app.post('/v1/introspection/%s/abort' % self.uuid)

        self.client_mock.call.assert_called_once_with({}, 'do_abort',
                                                      node_id=self.uuid,
                                                      token=None)
        self.assertEqual(404, res.status_code)
        data = json.loads(str(res.data.decode()))
        self.assertEqual(str(exc), data['error']['message'])

    def test_abort_failed(self):
        exc = utils.Error("Locked.", code=409)
        self.client_mock.call.side_effect = exc

        res = self.app.post('/v1/introspection/%s/abort' % self.uuid)

        self.client_mock.call.assert_called_once_with({}, 'do_abort',
                                                      node_id=self.uuid,
                                                      token=None)
        self.assertEqual(409, res.status_code)
        data = json.loads(res.data.decode())
        self.assertEqual(str(exc), data['error']['message'])


class GetStatusAPIBaseTest(BaseAPITest):
    def setUp(self):
        super(GetStatusAPIBaseTest, self).setUp()
        self.uuid2 = uuidutils.generate_uuid()
        self.finished_node = node_cache.NodeInfo(
            uuid=self.uuid,
            started_at=datetime.datetime(1, 1, 1),
            finished_at=datetime.datetime(1, 1, 2),
            error='boom',
            state=istate.States.error)
        self.finished_node.links = [
            {u'href': u'http://localhost/v1/introspection/%s' %
             self.finished_node.uuid,
             u'rel': u'self'},
        ]
        self.finished_node.status = {
            'finished': True,
            'state': self.finished_node._state,
            'started_at': self.finished_node.started_at.isoformat(),
            'finished_at': self.finished_node.finished_at.isoformat(),
            'error': self.finished_node.error,
            'uuid': self.finished_node.uuid,
            'links': self.finished_node.links
        }

        self.unfinished_node = node_cache.NodeInfo(
            uuid=self.uuid2,
            started_at=datetime.datetime(1, 1, 1),
            state=istate.States.processing)
        self.unfinished_node.links = [
            {u'href': u'http://localhost/v1/introspection/%s' %
             self.unfinished_node.uuid,
             u'rel': u'self'}
        ]
        finished_at = (self.unfinished_node.finished_at.isoformat()
                       if self.unfinished_node.finished_at else None)
        self.unfinished_node.status = {
            'finished': False,
            'state': self.unfinished_node._state,
            'started_at': self.unfinished_node.started_at.isoformat(),
            'finished_at': finished_at,
            'error': None,
            'uuid': self.unfinished_node.uuid,
            'links': self.unfinished_node.links
        }


@mock.patch.object(node_cache, 'get_node', autospec=True)
class TestApiGetStatus(GetStatusAPIBaseTest):
    def test_get_introspection_in_progress(self, get_mock):
        get_mock.return_value = self.unfinished_node
        res = self.app.get('/v1/introspection/%s' % self.uuid)
        self.assertEqual(200, res.status_code)
        self.assertEqual(self.unfinished_node.status,
                         json.loads(res.data.decode('utf-8')))

    def test_get_introspection_finished(self, get_mock):
        get_mock.return_value = self.finished_node
        res = self.app.get('/v1/introspection/%s' % self.uuid)
        self.assertEqual(200, res.status_code)
        self.assertEqual(self.finished_node.status,
                         json.loads(res.data.decode('utf-8')))


@mock.patch.object(node_cache, 'get_node_list', autospec=True)
class TestApiListStatus(GetStatusAPIBaseTest):

    def test_list_introspection(self, list_mock):
        list_mock.return_value = [self.finished_node, self.unfinished_node]
        res = self.app.get('/v1/introspection')
        self.assertEqual(200, res.status_code)
        statuses = json.loads(res.data.decode('utf-8')).get('introspection')

        self.assertEqual([self.finished_node.status,
                          self.unfinished_node.status], statuses)
        list_mock.assert_called_once_with(marker=None,
                                          limit=CONF.api_max_limit, state=None)

    def test_list_introspection_limit(self, list_mock):
        res = self.app.get('/v1/introspection?limit=1000')
        self.assertEqual(200, res.status_code)
        list_mock.assert_called_once_with(marker=None, limit=1000, state=None)

    def test_list_introspection_makrer(self, list_mock):
        res = self.app.get('/v1/introspection?marker=%s' %
                           self.finished_node.uuid)
        self.assertEqual(200, res.status_code)
        list_mock.assert_called_once_with(marker=self.finished_node.uuid,
                                          limit=CONF.api_max_limit, state=None)

    def test_list_introspection_state(self, list_mock):
        state = 'error'
        list_mock.return_value = [self.finished_node]
        res = self.app.get('/v1/introspection?state=%s' % state)
        self.assertEqual(200, res.status_code)
        statuses = json.loads(res.data.decode('utf-8')).get('introspection')

        self.assertEqual([self.finished_node.status], statuses)
        list_mock.assert_called_once_with(marker=None,
                                          limit=CONF.api_max_limit,
                                          state=state.split(','))

    def test_list_introspection_multiple_state(self, list_mock):
        state = 'error,processing'
        list_mock.return_value = [self.finished_node,
                                  self.unfinished_node]
        res = self.app.get('/v1/introspection?state=%s' % state)
        self.assertEqual(200, res.status_code)
        statuses = json.loads(res.data.decode('utf-8')).get('introspection')

        self.assertEqual([self.finished_node.status,
                          self.unfinished_node.status], statuses)
        list_mock.assert_called_once_with(marker=None,
                                          limit=CONF.api_max_limit,
                                          state=state.split(','))

    def test_list_introspection_invalid_state(self, list_mock):
        state = 'error,invalid'
        res = self.app.get('/v1/introspection?state=%s' % state)
        self.assertEqual(400, res.status_code)


class TestApiGetData(BaseAPITest):
    def setUp(self):
        super(TestApiGetData, self).setUp()
        self.introspection_data = {
            'ipmi_address': '1.2.3.4',
            'cpus': 2,
            'cpu_arch': 'x86_64',
            'memory_mb': 1024,
            'local_gb': 20,
            'interfaces': {
                'em1': {'mac': '11:22:33:44:55:66', 'ip': '1.2.0.1'},
            }
        }

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test_get_introspection_data_from_swift(self, swift_mock):
        CONF.set_override('store_data', 'swift', 'processing')
        swift_conn = swift_mock.return_value
        swift_conn.get_object.return_value = json.dumps(
            self.introspection_data)
        res = self.app.get('/v1/introspection/%s/data' % self.uuid)
        name = 'inspector_data-%s' % self.uuid
        swift_conn.get_object.assert_called_once_with(name)
        self.assertEqual(200, res.status_code)
        self.assertEqual(self.introspection_data,
                         json.loads(res.data.decode('utf-8')))

    @mock.patch.object(swift, 'SwiftAPI', autospec=True)
    def test_get_unprocessed_data_from_swift(self, swift_mock):
        CONF.set_override('store_data', 'swift', 'processing')
        swift_conn = swift_mock.return_value
        swift_conn.get_object.return_value = json.dumps(
            self.introspection_data)
        res = self.app.get('/v1/introspection/%s/data/unprocessed' % self.uuid)
        name = 'inspector_data-%s-UNPROCESSED' % self.uuid
        swift_conn.get_object.assert_called_once_with(name)
        self.assertEqual(200, res.status_code)
        self.assertEqual(self.introspection_data,
                         json.loads(res.data.decode('utf-8')))

    @mock.patch.object(intros_data_plugin, 'DatabaseStore',
                       autospec=True)
    def test_get_introspection_data_from_db(self, db_mock):
        CONF.set_override('store_data', 'database', 'processing')
        db_store = db_mock.return_value
        db_store.get.return_value = json.dumps(self.introspection_data)
        res = self.app.get('/v1/introspection/%s/data' % self.uuid)
        db_store.get.assert_called_once_with(self.uuid, processed=True,
                                             get_json=False)
        self.assertEqual(200, res.status_code)
        self.assertEqual(self.introspection_data,
                         json.loads(res.data.decode('utf-8')))

    def test_introspection_data_not_stored(self):
        CONF.set_override('store_data', 'none', 'processing')
        res = self.app.get('/v1/introspection/%s/data' % self.uuid)
        self.assertEqual(404, res.status_code)

    def test_unprocessed_data_not_stored(self):
        CONF.set_override('store_data', 'none', 'processing')
        res = self.app.get('/v1/introspection/%s/data/unprocessed' % self.uuid)
        self.assertEqual(404, res.status_code)

    @mock.patch.object(ir_utils, 'get_node', autospec=True)
    @mock.patch.object(main.process, 'get_introspection_data', autospec=True)
    def test_with_name(self, process_mock, get_mock):
        get_mock.return_value = mock.Mock(uuid=self.uuid)
        CONF.set_override('store_data', 'swift', 'processing')
        process_mock.return_value = json.dumps(self.introspection_data)
        res = self.app.get('/v1/introspection/name1/data')
        self.assertEqual(200, res.status_code)
        self.assertEqual(self.introspection_data,
                         json.loads(res.data.decode('utf-8')))
        get_mock.assert_called_once_with('name1', fields=['uuid'])


class TestApiReapply(BaseAPITest):

    def setUp(self):
        super(TestApiReapply, self).setUp()
        CONF.set_override('store_data', 'swift', 'processing')

    def test_api_ok(self):
        self.app.post('/v1/introspection/%s/data/unprocessed' %
                      self.uuid)
        self.client_mock.call.assert_called_once_with({}, 'do_reapply',
                                                      node_uuid=self.uuid,
                                                      data=None)

    def test_user_data(self):
        res = self.app.post('/v1/introspection/%s/data/unprocessed' %
                            self.uuid, data='some data')
        self.assertEqual(400, res.status_code)
        message = json.loads(res.data.decode())['error']['message']
        self.assertIn('Invalid data: expected a JSON object', message)
        self.assertFalse(self.client_mock.call.called)

    def test_user_data_valid(self):
        data = {"foo": "bar"}
        res = self.app.post('/v1/introspection/%s/data/unprocessed' %
                            self.uuid, data=json.dumps(data))
        self.assertEqual(202, res.status_code)
        self.client_mock.call.assert_called_once_with({}, 'do_reapply',
                                                      node_uuid=self.uuid,
                                                      data=data)

    def test_get_introspection_data_error(self):
        exc = utils.Error('The store is crashed', code=404)
        self.client_mock.call.side_effect = exc

        res = self.app.post('/v1/introspection/%s/data/unprocessed' %
                            self.uuid)

        self.assertEqual(404, res.status_code)
        message = json.loads(res.data.decode())['error']['message']
        self.assertEqual(str(exc), message)
        self.client_mock.call.assert_called_once_with({}, 'do_reapply',
                                                      node_uuid=self.uuid,
                                                      data=None)

    def test_generic_error(self):
        exc = utils.Error('Oops', code=400)
        self.client_mock.call.side_effect = exc

        res = self.app.post('/v1/introspection/%s/data/unprocessed' %
                            self.uuid)

        self.assertEqual(400, res.status_code)
        message = json.loads(res.data.decode())['error']['message']
        self.assertEqual(str(exc), message)
        self.client_mock.call.assert_called_once_with({}, 'do_reapply',
                                                      node_uuid=self.uuid,
                                                      data=None)

    @mock.patch.object(ir_utils, 'get_node', autospec=True)
    def test_reapply_with_node_name(self, get_mock):
        get_mock.return_value = mock.Mock(id=self.uuid)
        self.app.post('/v1/introspection/%s/data/unprocessed' %
                      'fake-node')
        self.client_mock.call.assert_called_once_with({}, 'do_reapply',
                                                      node_uuid=self.uuid,
                                                      data=None)
        get_mock.assert_called_once_with('fake-node', fields=['uuid'])


class TestApiRules(BaseAPITest):
    @mock.patch.object(rules, 'get_all', autospec=True)
    def test_get_all(self, get_all_mock):
        get_all_mock.return_value = [
            mock.Mock(spec=rules.IntrospectionRule,
                      **{'as_dict.return_value': {'uuid': 'foo'}}),
            mock.Mock(spec=rules.IntrospectionRule,
                      **{'as_dict.return_value': {'uuid': 'bar'}}),
        ]

        res = self.app.get('/v1/rules')
        self.assertEqual(200, res.status_code)
        self.assertEqual(
            {
                'rules': [{'uuid': 'foo',
                           'links': [
                               {'href': '/v1/rules/foo', 'rel': 'self'}
                           ]},
                          {'uuid': 'bar',
                           'links': [
                               {'href': '/v1/rules/bar', 'rel': 'self'}
                           ]}]
            },
            json.loads(res.data.decode('utf-8')))
        get_all_mock.assert_called_once_with()
        for m in get_all_mock.return_value:
            m.as_dict.assert_called_with(short=True)

    @mock.patch.object(rules, 'delete_all', autospec=True)
    def test_delete_all(self, delete_all_mock):
        res = self.app.delete('/v1/rules')
        self.assertEqual(204, res.status_code)
        delete_all_mock.assert_called_once_with()

    @mock.patch.object(rules, 'create', autospec=True)
    def test_create(self, create_mock):
        data = {'uuid': self.uuid,
                'conditions': 'cond',
                'actions': 'act'}
        exp = data.copy()
        exp['description'] = None
        create_mock.return_value = mock.Mock(spec=rules.IntrospectionRule,
                                             **{'as_dict.return_value': exp})

        res = self.app.post('/v1/rules', data=json.dumps(data))
        self.assertEqual(201, res.status_code)
        create_mock.assert_called_once_with(conditions_json='cond',
                                            actions_json='act',
                                            uuid=self.uuid,
                                            description=None,
                                            scope=None)
        self.assertEqual(exp, json.loads(res.data.decode('utf-8')))

    @mock.patch.object(rules, 'create', autospec=True)
    def test_create_api_less_1_6(self, create_mock):
        data = {'uuid': self.uuid,
                'conditions': 'cond',
                'actions': 'act'}
        exp = data.copy()
        exp['description'] = None
        create_mock.return_value = mock.Mock(spec=rules.IntrospectionRule,
                                             **{'as_dict.return_value': exp})

        headers = {conf_opts.VERSION_HEADER:
                   main._format_version((1, 5))}

        res = self.app.post('/v1/rules', data=json.dumps(data),
                            headers=headers)
        self.assertEqual(200, res.status_code)
        create_mock.assert_called_once_with(conditions_json='cond',
                                            actions_json='act',
                                            uuid=self.uuid,
                                            description=None,
                                            scope=None)
        self.assertEqual(exp, json.loads(res.data.decode('utf-8')))

    @mock.patch.object(rules, 'create', autospec=True)
    def test_create_bad_uuid(self, create_mock):
        data = {'uuid': 'foo',
                'conditions': 'cond',
                'actions': 'act'}

        res = self.app.post('/v1/rules', data=json.dumps(data))
        self.assertEqual(400, res.status_code)

    @mock.patch.object(rules, 'get', autospec=True)
    def test_get_one(self, get_mock):
        get_mock.return_value = mock.Mock(spec=rules.IntrospectionRule,
                                          **{'as_dict.return_value':
                                             {'uuid': 'foo'}})

        res = self.app.get('/v1/rules/' + self.uuid)
        self.assertEqual(200, res.status_code)
        self.assertEqual({'uuid': 'foo',
                          'links': [
                              {'href': '/v1/rules/foo', 'rel': 'self'}
                          ]},
                         json.loads(res.data.decode('utf-8')))
        get_mock.assert_called_once_with(self.uuid)
        get_mock.return_value.as_dict.assert_called_once_with(short=False)

    @mock.patch.object(rules, 'delete', autospec=True)
    def test_delete_one(self, delete_mock):
        res = self.app.delete('/v1/rules/' + self.uuid)
        self.assertEqual(204, res.status_code)
        delete_mock.assert_called_once_with(self.uuid)
        self.assertEqual('text/plain; charset=utf-8',
                         res.headers['content-type'])
        # NOTE(TheJulia): This should be able to be removed once
        # https://github.com/eventlet/eventlet/issues/746
        # is resolved and available. This ensures we don't
        # send out a content length header out which is the
        # less evil of the RFC7230 broken-ness.
        self.assertEqual('0', res.headers['Content-Length'])


class TestApiMisc(BaseAPITest):
    @mock.patch.object(node_cache, 'get_node', autospec=True)
    def test_404_expected(self, get_mock):
        get_mock.side_effect = utils.Error('boom', code=404)
        res = self.app.get('/v1/introspection/%s' % self.uuid)
        self.assertEqual(404, res.status_code)
        self.assertEqual('boom', _get_error(res))

    def test_404_unexpected(self):
        res = self.app.get('/v42')
        self.assertEqual(404, res.status_code)
        self.assertIn('not found', _get_error(res).lower())

    @mock.patch.object(node_cache, 'get_node', autospec=True)
    def test_500_with_debug(self, get_mock):
        CONF.set_override('debug', True)
        get_mock.side_effect = RuntimeError('boom')
        res = self.app.get('/v1/introspection/%s' % self.uuid)
        self.assertEqual(500, res.status_code)
        self.assertEqual('Internal server error (RuntimeError): boom',
                         _get_error(res))

    @mock.patch.object(node_cache, 'get_node', autospec=True)
    def test_500_without_debug(self, get_mock):
        CONF.set_override('debug', False)
        get_mock.side_effect = RuntimeError('boom')
        res = self.app.get('/v1/introspection/%s' % self.uuid)
        self.assertEqual(500, res.status_code)
        self.assertEqual('Internal server error',
                         _get_error(res))


class TestApiVersions(BaseAPITest):
    def _check_version_present(self, res):
        self.assertEqual('%d.%d' % main.MINIMUM_API_VERSION,
                         res.headers.get(conf_opts.MIN_VERSION_HEADER))
        self.assertEqual('%d.%d' % main.CURRENT_API_VERSION,
                         res.headers.get(conf_opts.MAX_VERSION_HEADER))

    def test_root_endpoint(self):
        res = self.app.get("/")
        self.assertEqual(200, res.status_code)
        self._check_version_present(res)
        data = res.data.decode('utf-8')
        json_data = json.loads(data)
        expected = {"versions": [{
            "status": "CURRENT", "id": '%s.%s' % main.CURRENT_API_VERSION,
            "links": [{
                "rel": "self",
                "href": ("http://localhost/v%s" %
                         main.CURRENT_API_VERSION[0])
            }]
        }]}
        self.assertEqual(expected, json_data)

    @mock.patch.object(main._app.url_map, "iter_rules", autospec=True)
    def test_version_endpoint(self, mock_rules):
        mock_rules.return_value = ["/v1/endpoint1", "/v1/endpoint2/<uuid>",
                                   "/v1/endpoint1/<name>/",
                                   "/v2/endpoint1", "/v1/endpoint3/",
                                   "/v1/endpoint2/<uuid>/subpoint"]
        endpoint = "/v1"
        res = self.app.get(endpoint)
        self.assertEqual(200, res.status_code)
        self._check_version_present(res)
        json_data = json.loads(res.data.decode('utf-8'))
        expected = {u'resources': [
            {
                u'name': u'endpoint1',
                u'links': [{
                    u'rel': u'self',
                    u'href': u'http://localhost/v1/endpoint1'}]
            },
            {
                u'name': u'endpoint3',
                u'links': [{
                    u'rel': u'self',
                    u'href': u'http://localhost/v1/endpoint3'}]
            },
        ]}
        self.assertEqual(expected, json_data)

    def test_version_endpoint_with_slash(self):
        endpoint = "/v1/"
        res = self.app.get(endpoint)
        self.assertEqual(200, res.status_code)
        self._check_version_present(res)

    def test_version_endpoint_invalid(self):
        endpoint = "/v-1"
        res = self.app.get(endpoint)
        self.assertEqual(404, res.status_code)

    def test_404_unexpected(self):
        # API version on unknown pages
        self._check_version_present(self.app.get('/v1/foobar'))

    @mock.patch.object(node_cache, 'get_node', autospec=True)
    def test_usual_requests(self, get_mock):
        get_mock.return_value = node_cache.NodeInfo(uuid=self.uuid,
                                                    started_at=42.0)
        # Successful
        self._check_version_present(
            self.app.post('/v1/introspection/%s' % self.uuid))
        # With error
        self._check_version_present(
            self.app.post('/v1/introspection/foobar'))

    def test_request_correct_version(self):
        headers = {conf_opts.VERSION_HEADER:
                   main._format_version(main.CURRENT_API_VERSION)}
        self._check_version_present(self.app.get('/', headers=headers))

    def test_request_correct_version_with_standard_singular_header(self):
        headers = {'OpenStack-API-Version': 'baremetal-introspection %s' %
                   main._format_version(main.CURRENT_API_VERSION)}
        self._check_version_present(self.app.get('/', headers=headers))

    def test_request_unsupported_version(self):
        bad_version = (main.CURRENT_API_VERSION[0],
                       main.CURRENT_API_VERSION[1] + 1)
        headers = {conf_opts.VERSION_HEADER:
                   main._format_version(bad_version)}
        res = self.app.get('/', headers=headers)
        self._check_version_present(res)
        self.assertEqual(406, res.status_code)
        error = _get_error(res)
        self.assertIn('%d.%d' % bad_version, error)
        self.assertIn('%d.%d' % main.MINIMUM_API_VERSION, error)
        self.assertIn('%d.%d' % main.CURRENT_API_VERSION, error)

    def test_request_latest_version(self):
        headers = {conf_opts.VERSION_HEADER: 'latest'}
        res = self.app.get('/', headers=headers)
        self.assertEqual(200, res.status_code)
        self._check_version_present(res)

    def test_request_latest_version_with_standard_singular_header(self):
        headers = {'OpenStack-API-Version': 'baremetal-introspection latest'}
        res = self.app.get('/', headers=headers)
        self.assertEqual(200, res.status_code)
        self._check_version_present(res)


class TestPlugins(unittest.TestCase):
    @mock.patch.object(example_plugin.ExampleProcessingHook,
                       'before_processing', autospec=True)
    @mock.patch.object(example_plugin.ExampleProcessingHook,
                       'before_update', autospec=True)
    def test_hook(self, mock_post, mock_pre):
        CONF.set_override('processing_hooks', 'example', 'processing')
        mgr = plugins_base.processing_hooks_manager()
        mgr.map_method('before_processing', 'introspection_data')
        mock_pre.assert_called_once_with(mock.ANY, 'introspection_data')
        mgr.map_method('before_update', 'node_info', {})
        mock_post.assert_called_once_with(mock.ANY, 'node_info', {})

    def test_manager_is_cached(self):
        self.assertIs(plugins_base.processing_hooks_manager(),
                      plugins_base.processing_hooks_manager())


class TestTopic(test_base.BaseTest):
    def setUp(self):
        super(TestTopic, self).setUp()
        self.transport_mock = self.useFixture(
            fixtures.MockPatchObject(rpc, 'TRANSPORT',
                                     autospec=True)).mock
        self.target_mock = self.useFixture(
            fixtures.MockPatchObject(rpc.messaging, 'Target',
                                     autospec=True)).mock
        self.rpcclient_mock = self.useFixture(
            fixtures.MockPatchObject(rpc.messaging, 'RPCClient',
                                     autospec=True)).mock
        CONF.set_override('host', 'a-host')

    def test_get_client_compat_standalone(self):
        main.get_client_compat()
        self.target_mock.assert_called_with(topic='ironic_inspector.conductor',
                                            server='a-host', version=mock.ANY)

    @mock.patch.object(main, 'get_random_topic', autospec=True)
    def test_get_client_compat_non_standalone(self, mock_get_topic):
        CONF.set_override('host', 'a-host')
        CONF.set_override('standalone', False)
        mock_get_topic.return_value = 'hello'
        main.get_client_compat()
        self.target_mock.assert_called_with(
            topic='hello', version=mock.ANY)

    @mock.patch.object(coordination, 'get_coordinator', autospec=True)
    def test_get_random_topic(self, mock_get_coordinator):
        mock_coordinator = mock.Mock(spec=['get_members'])
        members = [('ironic_inspector.conductor.host%s' % i).encode('ascii')
                   for i in range(5)]
        topics = [('ironic_inspector.conductor.host%s' % i) for i in range(5)]
        mock_coordinator.get_members.return_value = set(members)
        mock_get_coordinator.return_value = mock_coordinator
        for i in range(10):
            topic = main.get_random_topic()
            self.assertIn(topic, topics)

    @mock.patch.object(coordination, 'get_coordinator', autospec=True)
    def test_get_random_topic_host_with_domain(self, mock_get_coordinator):
        mock_coordinator = mock.Mock(spec=['get_members'])
        members = ['ironic_inspector.conductor.'
                   'local.domain'.encode('ascii')]
        mock_coordinator.get_members.return_value = set(members)
        mock_get_coordinator.return_value = mock_coordinator
        topic = main.get_random_topic()
        self.assertEqual(topic, 'ironic_inspector.conductor.local.domain')

    @mock.patch.object(coordination, 'get_coordinator', autospec=True)
    def test_get_random_topic_host_bypass_invalid(self, mock_get_coordinator):
        mock_coordinator = mock.Mock(spec=['get_members'])
        members = ['this_should_not_happen'.encode('ascii')]
        mock_coordinator.get_members.return_value = set(members)
        mock_get_coordinator.return_value = mock_coordinator
        self.assertRaisesRegex(utils.NoAvailableConductor,
                               'No available conductor',
                               main.get_random_topic)

    @mock.patch.object(coordination, 'get_coordinator', autospec=True)
    def test_get_random_topic_no_member(self, mock_get_coordinator):
        mock_coordinator = mock.Mock(spec=['get_members'])
        mock_coordinator.get_members.return_value = set()
        mock_get_coordinator.return_value = mock_coordinator
        self.assertRaises(utils.NoAvailableConductor, main.get_random_topic)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gather root device hint from recognized block devices."""

from ironic_inspector.plugins import base
from ironic_inspector import utils


LOG = utils.getProcessingLogger(__name__)


class RaidDeviceDetection(base.ProcessingHook):
    """Processing hook for learning the root device after RAID creation.

    The plugin can figure out the root device in 2 runs. First, it saves the
    discovered block device serials in node.extra. The second run will check
    the difference between the recently discovered block devices and the
    previously saved ones. After saving the root device in node.properties, it
    will delete the temporarily saved block device serials in node.extra.

    This way, it helps to figure out the root device hint in cases when
    otherwise Ironic doesn't have enough information to do so. Such a usecase
    is DRAC RAID configuration where the BMC doesn't provide any useful
    information about the created RAID disks. Using this plugin immediately
    before and after creating the root RAID device will solve the issue of root
    device hints.

    In cases where there's no RAID volume on the node, the standard plugin will
    fail due to the missing local_gb value. This plugin fakes the missing
    value, until it's corrected during later runs. Note, that for this to work
    the plugin needs to take precedence over the standard plugin.
    """

    def _get_serials(self, data):
        if 'inventory' in data:
            return [x['serial'] for x in data['inventory'].get('disks', ())
                    if x.get('serial')]
        elif 'block_devices' in data:
            return data['block_devices'].get('serials', ())

    def before_processing(self, introspection_data, **kwargs):
        """Adds fake local_gb value if it's missing from introspection_data."""
        if not introspection_data.get('local_gb'):
            LOG.info('No volume is found on the node. Adding a fake '
                     'value for "local_gb"', data=introspection_data)
            introspection_data['local_gb'] = 1

    def before_update(self, introspection_data, node_info, **kwargs):
        current_devices = self._get_serials(introspection_data)
        if not current_devices:
            LOG.warning('No block device was received from ramdisk',
                        node_info=node_info, data=introspection_data)
            return

        node = node_info.node()

        if 'root_device' in node.properties:
            LOG.info('Root device is already known for the node',
                     node_info=node_info, data=introspection_data)
            return

        if 'block_devices' in node.extra:
            # Compare previously discovered devices with the current ones
            previous_devices = node.extra['block_devices']['serials']
            new_devices = [device for device in current_devices
                           if device not in previous_devices]

            if len(new_devices) > 1:
                LOG.warning('Root device cannot be identified because '
                            'multiple new devices were found',
                            node_info=node_info, data=introspection_data)
                return
            elif len(new_devices) == 0:
                LOG.warning('No new devices were found',
                            node_info=node_info, data=introspection_data)
                return

            node_info.patch([
                {'op': 'remove',
                 'path': '/extra/block_devices'},
                {'op': 'add',
                 'path': '/properties/root_device',
                 'value': {'serial': new_devices[0]}}
            ])

        else:
            # No previously discovered devices - save the inspector block
            # devices in node.extra
            node_info.patch([{'op': 'add',
                              'path': '/extra/block_devices',
                              'value': {'serials': current_devices}}])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Standard plugins for rules API."""

import operator
import re

import netaddr
from openstack import exceptions as os_exc

from ironic_inspector.common.i18n import _
from ironic_inspector.plugins import base
from ironic_inspector import utils


LOG = utils.getProcessingLogger(__name__)


def coerce(value, expected):
    if isinstance(expected, float):
        return float(value)
    elif isinstance(expected, int):
        return int(value)
    else:
        return value


class SimpleCondition(base.RuleConditionPlugin):
    op = None

    def check(self, node_info, field, params, **kwargs):
        value = params['value']
        return self.op(coerce(field, value), value)


class EqCondition(SimpleCondition):
    op = operator.eq


class LtCondition(SimpleCondition):
    op = operator.lt


class GtCondition(SimpleCondition):
    op = operator.gt


class LeCondition(SimpleCondition):
    op = operator.le


class GeCondition(SimpleCondition):
    op = operator.ge


class NeCondition(SimpleCondition):
    op = operator.ne


class EmptyCondition(base.RuleConditionPlugin):
    REQUIRED_PARAMS = set()
    ALLOW_NONE = True

    def check(self, node_info, field, params, **kwargs):
        return field in ('', None, [], {})


class NetCondition(base.RuleConditionPlugin):
    def validate(self, params, **kwargs):
        super(NetCondition, self).validate(params, **kwargs)
        # Make sure it does not raise
        try:
            netaddr.IPNetwork(params['value'])
        except netaddr.AddrFormatError as exc:
            raise ValueError('invalid value: %s' % exc)

    def check(self, node_info, field, params, **kwargs):
        network = netaddr.IPNetwork(params['value'])
        return netaddr.IPAddress(field) in network


class ReCondition(base.RuleConditionPlugin):
    def validate(self, params, **kwargs):
        try:
            re.compile(params['value'])
        except re.error as exc:
            raise ValueError(_('invalid regular expression: %s') % exc)


class MatchesCondition(ReCondition):
    def check(self, node_info, field, params, **kwargs):
        regexp = params['value']
        if regexp[-1] != '$':
            regexp += '$'
        return re.match(regexp, str(field)) is not None


class ContainsCondition(ReCondition):
    def check(self, node_info, field, params, **kwargs):
        return re.search(params['value'], str(field)) is not None


class FailAction(base.RuleActionPlugin):
    REQUIRED_PARAMS = {'message'}

    def apply(self, node_info, params, **kwargs):
        raise utils.Error(params['message'], node_info=node_info)


class SetAttributeAction(base.RuleActionPlugin):
    # NOTE(iurygregory): set as optional to accept None as value, check
    # that the key 'value' is present, otherwise will raise ValueError.
    OPTIONAL_PARAMS = {'value', 'reset_interfaces'}
    REQUIRED_PARAMS = {'path'}
    # TODO(dtantsur): proper validation of path

    FORMATTED_PARAMS = ['value']

    def apply(self, node_info, params, **kwargs):
        kwargs = {}
        if (params['path'].strip('/') == 'driver'
                and ('reset_interfaces' not in params
                     or params['reset_interfaces'])):
            # Using at least Rocky
            kwargs['reset_interfaces'] = True

        try:
            node_info.patch([{'op': 'add', 'path': params['path'],
                              'value': params['value']}], **kwargs)
        except (TypeError, os_exc.SDKException):
            # TODO(dtantsur): remove support for old ironicclient and Queens
            if 'reset_interfaces' in params:
                # An explicit request, report an error.
                raise utils.Error(
                    _('Cannot pass reset_interfaces to set-attribute, '
                      'requires API 1.46 and ironicclient >= 2.5.0'))

            LOG.warning('Not passing reset_interfaces to Ironic, since '
                        ' API 1.46 and/or ironicclient >= 2.5.0 are '
                        'not available', node_info=node_info)
            node_info.patch([{'op': 'add', 'path': params['path'],
                              'value': params['value']}])

    def validate(self, params, **kwargs):
        if 'value' in params:
            super(base.RuleActionPlugin, self).validate(params, **kwargs)
        else:
            msg = _('missing required parameter(s): value')
            raise ValueError(msg)


class SetCapabilityAction(base.RuleActionPlugin):
    REQUIRED_PARAMS = {'name'}
    OPTIONAL_PARAMS = {'value'}

    FORMATTED_PARAMS = ['value']

    def apply(self, node_info, params, **kwargs):
        node_info.update_capabilities(
            **{params['name']: params.get('value')})


class ExtendAttributeAction(base.RuleActionPlugin):
    REQUIRED_PARAMS = {'path', 'value'}
    OPTIONAL_PARAMS = {'unique'}
    # TODO(dtantsur): proper validation of path

    FORMATTED_PARAMS = ['value']

    def apply(self, node_info, params, **kwargs):
        def _replace(values):
            value = params['value']
            if not params.get('unique') or value not in values:
                values.append(value)
            return values

        node_info.replace_field(params['path'], _replace, default=[])


class AddTraitAction(base.RuleActionPlugin):
    REQUIRED_PARAMS = {'name'}

    def apply(self, node_info, params, **kwargs):
        node_info.add_trait(params['name'])


class RemoveTraitAction(base.RuleActionPlugin):
    REQUIRED_PARAMS = {'name'}

    def apply(self, node_info, params, **kwargs):
        node_info.remove_trait(params['name'])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTE(milan) the filter design relies on the hostdir[1] being in exclusive
# inspector control. The hostdir should be considered a private cache directory
# of inspector that dnsmasq has read access to and polls updates from, through
# the inotify facility.
#
# [1] see the --dhcp-hostsdir option description in
#     http://www.thekelleys.org.uk/dnsmasq/docs/dnsmasq-man.html


try:
    import errno
except ImportError:
    import os.errno as errno
import fcntl
import os
import time

from keystoneauth1 import exceptions as ksa_exc
from openstack import exceptions as os_exc
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from ironic_inspector.common import ironic as ir_utils
from ironic_inspector import node_cache
from ironic_inspector.pxe_filter import base as pxe_filter

CONF = cfg.CONF
LOG = log.getLogger(__name__)

_EXCLUSIVE_WRITE_ATTEMPTS = 10
_EXCLUSIVE_WRITE_ATTEMPTS_DELAY = 0.01

_ROOTWRAP_COMMAND = 'sudo ironic-inspector-rootwrap {rootwrap_config!s}'
_MAC_DENY_LEN = len('ff:ff:ff:ff:ff:ff,ignore\n')
_MAC_ALLOW_LEN = len('ff:ff:ff:ff:ff:ff\n')
_UNKNOWN_HOSTS_FILE = 'unknown_hosts_filter'
_DENY_UNKNOWN_HOSTS = '*:*:*:*:*:*,ignore\n'
_ALLOW_UNKNOWN_HOSTS = '*:*:*:*:*:*\n'


def _should_enable_unknown_hosts():
    """Check whether we should enable DHCP for unknown hosts

    We add unknown hosts to the deny list unless one or more nodes are on
    introspection, or the node_not_found_hook is not set.
    """
    return (node_cache.introspection_active() or
            CONF.processing.node_not_found_hook is not None)


class DnsmasqFilter(pxe_filter.BaseFilter):
    """The dnsmasq PXE filter driver.

    A pxe filter driver implementation that controls access to dnsmasq
    through amending its configuration.
    """

    def reset(self):
        """Stop dnsmasq and upcall reset."""
        _execute(CONF.dnsmasq_pxe_filter.dnsmasq_stop_command,
                 ignore_errors=True)
        super(DnsmasqFilter, self).reset()

    def _sync(self, ironic):
        """Sync the inspector, ironic and dnsmasq state. Locked.

        :raises: IOError, OSError.
        :returns: None.
        """
        LOG.debug('Syncing the driver')
        timestamp_start = timeutils.utcnow()

        try:
            # active_macs are the MACs for which introspection is active
            active_macs = pxe_filter.get_active_macs(ironic)

            # ironic_macs are all the MACs know to ironic (all ironic ports)
            ironic_macs = pxe_filter.get_ironic_macs(ironic)
        except (os_exc.SDKException, ksa_exc.ConnectFailure):
            LOG.exception(
                "Could not list ironic ports, can not sync dnsmasq PXE filter")
            return

        denylist, allowlist = _get_deny_allow_lists()
        # removedlist are the MACs that are in either in allow or denylist,
        # but not kept in ironic (ironic_macs) any more
        removedlist = denylist.union(allowlist).difference(ironic_macs)

        # Add active MACs that are not already in the allowlist
        for mac in active_macs.difference(allowlist):
            _add_mac_to_allowlist(mac)
        # Add any ironic MACs that is not active for introspection to the
        # deny list unless it is already present
        for mac in ironic_macs.difference(denylist.union(active_macs)):
            _add_mac_to_denylist(mac)

        # Allow or deny unknown hosts and MACs not kept in ironic
        # NOTE(hjensas): Treat unknown hosts and MACs not kept in ironic the
        # same. Neither should boot the inspection image unless introspection
        # is active. Deleted MACs must be added to the allow list when
        # introspection is active in case the host is re-enrolled.
        _configure_unknown_hosts()
        _configure_removedlist(removedlist)

        timestamp_end = timeutils.utcnow()
        LOG.debug('The dnsmasq PXE filter was synchronized (took %s)',
                  timestamp_end - timestamp_start)

    @pxe_filter.locked_driver_event(pxe_filter.Events.sync)
    def sync(self, ironic):
        """Sync dnsmasq configuration with current Ironic&Inspector state.

        Polls all ironic ports. Those being inspected, the active ones, are
        added to the allow list while the rest are added to the deny list in
        the dnsmasq configuration.

        :param ironic: an ironic client instance.
        :raises: OSError, IOError.
        :returns: None.
        """
        self._sync(ironic)

    @pxe_filter.locked_driver_event(pxe_filter.Events.initialize)
    def init_filter(self):
        """Performs an initial sync with ironic and starts dnsmasq.

        The initial _sync() call reduces the chances dnsmasq might lose
        some inotify deny list events by prefetching the list before dnsmasq
        is started.

        :raises: OSError, IOError.
        :returns: None.
        """
        _purge_dhcp_hostsdir()
        ironic = ir_utils.get_client()
        self._sync(ironic)
        _execute(CONF.dnsmasq_pxe_filter.dnsmasq_start_command)
        LOG.info('The dnsmasq PXE filter was initialized')


def _purge_dhcp_hostsdir():
    """Remove all the DHCP hosts files.

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid.
             IOError in case of non-writable file or a record not being a file.
    :returns: None.
    """
    dhcp_hostsdir = CONF.dnsmasq_pxe_filter.dhcp_hostsdir
    if not CONF.dnsmasq_pxe_filter.purge_dhcp_hostsdir:
        LOG.debug('Not purging %s; disabled in configuration.', dhcp_hostsdir)
        return

    LOG.debug('Purging %s', dhcp_hostsdir)
    for mac in os.listdir(dhcp_hostsdir):
        path = os.path.join(dhcp_hostsdir, mac)
        # NOTE(milan) relying on a failure here aborting the init_filter() call
        os.remove(path)
        LOG.debug('Removed %s', path)


def _get_deny_allow_lists():
    """Get addresses currently denied by dnsmasq.

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid.
    :returns: a set of MACs currently denied by dnsmasq.
    """
    hostsdir = CONF.dnsmasq_pxe_filter.dhcp_hostsdir
    # MACs in the allow list lack the ,ignore directive
    denylist = set()
    allowlist = set()
    for mac in os.listdir(hostsdir):
        if os.stat(os.path.join(hostsdir, mac)).st_size == _MAC_DENY_LEN:
            denylist.add(mac)
        if os.stat(os.path.join(hostsdir, mac)).st_size == _MAC_ALLOW_LEN:
            allowlist.add(mac)

    return denylist, allowlist


def _exclusive_write_or_pass(path, buf):
    """Write exclusively or pass if path locked.

    The intention is to be able to run multiple instances of the filter on the
    same node in multiple inspector processes.

    :param path: where to write to
    :param buf: the content to write
    :raises: FileNotFoundError, IOError
    :returns: True if the write was successful.
    """
    # NOTE(milan) line-buffering enforced to ensure dnsmasq record update
    # through inotify, which reacts on f.close()
    attempts = _EXCLUSIVE_WRITE_ATTEMPTS
    with open(path, 'w', 1) as f:
        while attempts:
            try:
                fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
                f.write(buf)
                # Go ahead and flush the data now instead of waiting until
                # after the automatic flush with the file close after the
                # file lock is released.
                f.flush()
                return True
            except IOError as e:
                if e.errno == errno.EWOULDBLOCK:
                    LOG.debug('%s locked; will try again (later)', path)
                    attempts -= 1
                    time.sleep(_EXCLUSIVE_WRITE_ATTEMPTS_DELAY)
                    continue
                raise
            finally:
                fcntl.flock(f, fcntl.LOCK_UN)
    LOG.debug('Failed to write the exclusively-locked path: %(path)s for '
              '%(attempts)s times', {'attempts': _EXCLUSIVE_WRITE_ATTEMPTS,
                                     'path': path})
    return False


def _configure_removedlist(macs):
    """Manages a dhcp_hostsdir allow/deny record for removed macs

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid,
    :returns: None.
    """

    hostsdir = CONF.dnsmasq_pxe_filter.dhcp_hostsdir

    if (_should_enable_unknown_hosts()
            and not CONF.pxe_filter.deny_unknown_macs):
        for mac in macs:
            if os.stat(os.path.join(hostsdir, mac)).st_size != _MAC_ALLOW_LEN:
                _add_mac_to_allowlist(mac)
    else:
        for mac in macs:
            if os.stat(os.path.join(hostsdir, mac)).st_size != _MAC_DENY_LEN:
                _add_mac_to_denylist(mac)


def _configure_unknown_hosts():
    """Manages a dhcp_hostsdir allow/deny record for unknown macs.

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid,
             IOError in case the dhcp host unknown file isn't writable.
    :returns: None.
    """
    path = os.path.join(CONF.dnsmasq_pxe_filter.dhcp_hostsdir,
                        _UNKNOWN_HOSTS_FILE)

    if (_should_enable_unknown_hosts()
            and not CONF.pxe_filter.deny_unknown_macs):
        wildcard_filter = _ALLOW_UNKNOWN_HOSTS
        log_wildcard_filter = 'allow'
    else:
        wildcard_filter = _DENY_UNKNOWN_HOSTS
        log_wildcard_filter = 'deny'

    # Don't update if unknown hosts are already in the deny/allow-list
    try:
        if os.stat(path).st_size == len(wildcard_filter):
            return
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise

    if _exclusive_write_or_pass(path, '%s' % wildcard_filter):
        LOG.debug('A %s record for all unknown hosts using wildcard mac '
                  'created', log_wildcard_filter)
    else:
        LOG.warning('Failed to %s unknown hosts using wildcard mac; '
                    'retrying next periodic sync time', log_wildcard_filter)


def _add_mac_to_denylist(mac):
    """Creates a dhcp_hostsdir deny record for the MAC.

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid,
             IOError in case the dhcp host MAC file isn't writable.
    :returns: None.
    """
    path = os.path.join(CONF.dnsmasq_pxe_filter.dhcp_hostsdir, mac)
    if _exclusive_write_or_pass(path, '%s,ignore\n' % mac):
        LOG.debug('MAC %s added to the deny list', mac)
    else:
        LOG.warning('Failed to add MAC %s to the deny list; retrying next '
                    'periodic sync time', mac)


def _add_mac_to_allowlist(mac):
    """Update the dhcp_hostsdir record for the MAC adding it to allow list

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid,
             IOError in case the dhcp host MAC file isn't writable.
    :returns: None.
    """
    path = os.path.join(CONF.dnsmasq_pxe_filter.dhcp_hostsdir, mac)
    # remove the ,ignore directive
    if _exclusive_write_or_pass(path, '%s\n' % mac):
        LOG.debug('MAC %s removed from the deny list', mac)
    else:
        LOG.warning('Failed to remove MAC %s from the deny list; retrying '
                    'next periodic sync time', mac)


def _execute(cmd=None, ignore_errors=False):
    # e.g: '/bin/kill $(cat /var/run/dnsmasq.pid)'
    if not cmd:
        return

    helper = _ROOTWRAP_COMMAND.format(rootwrap_config=CONF.rootwrap_config)
    processutils.execute(cmd, run_as_root=True, root_helper=helper, shell=True,
                         check_exit_code=not ignore_errors)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest import mock
import uuid

import fixtures
from keystoneauth1.fixture import v3 as v3_token
from keystonemiddleware import auth_token
from oslo_config import cfg
from oslo_context import context as oslo_context
import oslo_messaging as messaging
from oslo_utils import uuidutils

from ironic_inspector.common import rpc
import ironic_inspector.conf
from ironic_inspector import introspection_state as istate
from ironic_inspector import main
from ironic_inspector import node_cache
from ironic_inspector.test import base as test_base

CONF = ironic_inspector.conf.CONF


# Tokens for RBAC policy tests
ADMIN_TOKEN = uuid.uuid4().hex
admin_context = oslo_context.RequestContext(
    user_id=ADMIN_TOKEN,
    roles=['admin', 'member', 'reader'],
)

MANAGER_TOKEN = uuid.uuid4().hex
manager_context = oslo_context.RequestContext(
    user_id=MANAGER_TOKEN,
    roles=['manager', 'member', 'reader'],
)

MEMBER_TOKEN = uuid.uuid4().hex
member_context = oslo_context.RequestContext(
    user_id=MEMBER_TOKEN,
    roles=['member', 'reader'],
)

READER_TOKEN = uuid.uuid4().hex
reader_context = oslo_context.RequestContext(
    user_id=READER_TOKEN,
    roles=['reader'],
)

NO_ROLE_TOKEN = uuid.uuid4().hex
no_role_context = oslo_context.RequestContext(
    user_id=READER_TOKEN,
    roles=[],
)

SERVICE_TOKEN = uuid.uuid4().hex
service_context = oslo_context.RequestContext(
    user_id=SERVICE_TOKEN,
    roles=['service']
)

# Tokens for deprecated policy tests
BM_ADMIN_TOKEN = uuid.uuid4().hex
bm_admin_context = oslo_context.RequestContext(
    user_id=BM_ADMIN_TOKEN,
    roles=['baremetal_admin'],
)
BM_OBSERVER_TOKEN = uuid.uuid4().hex
bm_observer_context = oslo_context.RequestContext(
    user_id=BM_OBSERVER_TOKEN,
    roles=['baremetal_observer'],
)


USERS = {
    ADMIN_TOKEN: admin_context.to_dict(),
    MANAGER_TOKEN: manager_context.to_dict(),
    SERVICE_TOKEN: service_context.to_dict(),
    MEMBER_TOKEN: member_context.to_dict(),
    READER_TOKEN: reader_context.to_dict(),
    NO_ROLE_TOKEN: no_role_context.to_dict(),
    BM_ADMIN_TOKEN: bm_admin_context.to_dict(),
    BM_OBSERVER_TOKEN: bm_observer_context.to_dict(),
}


class BasePolicyTest(test_base.BaseTest):

    system_scope = False

    def init_app(self):
        CONF.set_override('auth_strategy', 'keystone')
        main._app.testing = True
        self.app = main.get_app().test_client()

    def setUp(self):
        super(BasePolicyTest, self).setUp()
        self.init_app()
        self.uuid = uuidutils.generate_uuid()
        self.rpc_get_client_mock = self.useFixture(
            fixtures.MockPatchObject(rpc, 'get_client', autospec=True)).mock
        self.client_mock = mock.MagicMock(spec=messaging.RPCClient)
        self.rpc_get_client_mock.return_value = self.client_mock

        self.fake_token = None
        mock_auth = mock.patch.object(
            auth_token.AuthProtocol, 'process_request',
            autospec=True)
        self.mock_auth = mock_auth.start()
        self.addCleanup(mock_auth.stop)
        self.mock_auth.side_effect = self._fake_process_request

        mock_get = mock.patch.object(node_cache, 'get_node', autospec=True)
        get = mock_get.start()
        self.addCleanup(mock_get.stop)
        get.return_value = node_cache.NodeInfo(
            uuid=self.uuid,
            started_at=datetime.datetime(1, 1, 1),
            state=istate.States.processing)

    def _fake_process_request(self, request, meow):
        if self.fake_token:
            request.user_token_valid = True
            request.user_token = True
            # is this right?!?
            request.token_info = self.fake_token
            request.auth_token = v3_token.Token(
                user_id=self.fake_token['user'])
        else:
            # Because of this, the user will always get a 403 in testing, even
            # if the API would normally return a 401 if a token is valid
            request.user_token_valid = False

    def set_token(self, token):
        self.fake_token = USERS[token]
        headers = {
            'X-Auth-Token': token,
            'X-Roles': ','.join(self.fake_token['roles'])
        }
        if self.system_scope:
            headers['OpenStack-System-Scope'] = 'all'
        return headers

    def assert_status(self, status_code, token, request_func, path, data=None):
        headers = self.set_token(token)
        res = request_func(path, headers=headers, data=data)
        self.assertEqual(status_code, res.status_code)


class TestACLDeprecated(BasePolicyTest):

    def setUp(self):
        super(TestACLDeprecated, self).setUp()
        cfg.CONF.set_override('enforce_scope', False, group='oslo_policy')
        cfg.CONF.set_override('enforce_new_defaults', False,
                              group='oslo_policy')

    def test_root_baremetal_admin(self):
        self.assert_status(200, BM_ADMIN_TOKEN, self.app.get, '/')
        self.assert_status(200, BM_ADMIN_TOKEN, self.app.get, '/v1')

    def test_root_baremetal_observer(self):
        self.assert_status(200, BM_OBSERVER_TOKEN, self.app.get, '/')
        self.assert_status(200, BM_OBSERVER_TOKEN, self.app.get, '/v1')

    def test_root_system_no_role(self):
        self.assert_status(200, NO_ROLE_TOKEN, self.app.get, '/')
        self.assert_status(200, NO_ROLE_TOKEN, self.app.get, '/v1')

    def test_introspect_baremetal_admin(self):
        self.assert_status(202, BM_ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s' % self.uuid)

    def test_introspect_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_baremetal_admin(self):
        self.assert_status(202, BM_ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_status_baremetal_admin(self):
        self.assert_status(200, BM_ADMIN_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_status_baremetal_observer(self):
        self.assert_status(200, BM_OBSERVER_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_list_baremetal_admin(self):
        self.assert_status(200, BM_ADMIN_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_list_baremetal_observer(self):
        self.assert_status(200, BM_OBSERVER_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_data_baremetal_admin(self):
        self.assert_status(404, BM_ADMIN_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_unprocessed_baremetal_admin(self):
        self.assert_status(400, BM_ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_data_unprocessed_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_rule_list_baremetal_admin(self):
        self.assert_status(200, BM_ADMIN_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_list_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_get_baremetal_admin(self):
        self.assert_status(404, BM_ADMIN_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_get_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_delete_all_baremetal_admin(self):
        self.assert_status(204, BM_ADMIN_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_all_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_baremetal_admin(self):
        self.assert_status(404, BM_ADMIN_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_delete_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_create_baremetal_admin(self):
        self.assert_status(500, BM_ADMIN_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })

    def test_rule_create_baremetal_observer(self):
        self.assert_status(403, BM_OBSERVER_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })


class TestRBACScoped(BasePolicyTest):

    system_scope = True

    def setUp(self):
        super(TestRBACScoped, self).setUp()
        cfg.CONF.set_override('enforce_scope', True, group='oslo_policy')
        cfg.CONF.set_override('enforce_new_defaults', True,
                              group='oslo_policy')

    def test_root_system_admin(self):
        self.assert_status(200, ADMIN_TOKEN, self.app.get, '/')
        self.assert_status(200, ADMIN_TOKEN, self.app.get, '/v1')

    def test_root_system_service(self):
        self.assert_status(200, SERVICE_TOKEN, self.app.get, '/')
        self.assert_status(200, SERVICE_TOKEN, self.app.get, '/v1')

    def test_root_system_member(self):
        self.assert_status(200, MEMBER_TOKEN, self.app.get, '/')
        self.assert_status(200, MEMBER_TOKEN, self.app.get, '/v1')

    def test_root_system_reader(self):
        self.assert_status(200, READER_TOKEN, self.app.get, '/')
        self.assert_status(200, READER_TOKEN, self.app.get, '/v1')

    def test_root_system_no_role(self):
        self.assert_status(200, NO_ROLE_TOKEN, self.app.get, '/')
        self.assert_status(200, NO_ROLE_TOKEN, self.app.get, '/v1')

    def test_introspect_system_admin(self):
        self.assert_status(202, ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s' % self.uuid)

    def test_introspect_system_service(self):
        self.assert_status(202, SERVICE_TOKEN, self.app.post,
                           '/v1/introspection/%s' % self.uuid)

    def test_introspect_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.post,
                           '/v1/introspection/%s' % self.uuid)

    def test_introspect_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_system_admin(self):
        self.assert_status(202, ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_system_service(self):
        self.assert_status(202, SERVICE_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_status_system_admin(self):
        self.assert_status(200, ADMIN_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_status_system_service(self):
        self.assert_status(200, SERVICE_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_status_system_member(self):
        self.assert_status(200, MEMBER_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_status_system_reader(self):
        self.assert_status(200, READER_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_list_system_admin(self):
        self.assert_status(200, ADMIN_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_list_system_service(self):
        self.assert_status(200, SERVICE_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_list_system_member(self):
        self.assert_status(200, MEMBER_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_list_system_reader(self):
        self.assert_status(200, READER_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_data_system_admin(self):
        self.assert_status(404, ADMIN_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_system_service(self):
        self.assert_status(404, SERVICE_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_unprocessed_system_admin(self):
        self.assert_status(400, ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_data_unprocessed_system_service(self):
        self.assert_status(400, SERVICE_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_data_unprocessed_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_data_unprocessed_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_rule_list_system_admin(self):
        self.assert_status(200, ADMIN_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_list_system_service(self):
        self.assert_status(200, SERVICE_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_list_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_list_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_get_system_admin(self):
        self.assert_status(404, ADMIN_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_get_system_service(self):
        self.assert_status(404, SERVICE_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_get_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_get_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_delete_all_system_admin(self):
        self.assert_status(204, ADMIN_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_all_system_service(self):
        self.assert_status(204, SERVICE_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_all_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_all_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_system_admin(self):
        self.assert_status(404, ADMIN_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_delete_system_service(self):
        self.assert_status(404, SERVICE_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_delete_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_delete_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_create_system_admin(self):
        self.assert_status(500, ADMIN_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })

    def test_rule_create_system_service(self):
        self.assert_status(500, SERVICE_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })

    def test_rule_create_system_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })

    def test_rule_create_system_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })


class TestRBACProjectScope(BasePolicyTest):

    system_scope = False

    def setUp(self):
        super(TestRBACProjectScope, self).setUp()
        cfg.CONF.set_override('enforce_scope', True, group='oslo_policy')
        cfg.CONF.set_override('enforce_new_defaults', True,
                              group='oslo_policy')

    def test_root_project_admin(self):
        self.assert_status(200, ADMIN_TOKEN, self.app.get, '/')
        self.assert_status(200, ADMIN_TOKEN, self.app.get, '/v1')

    def test_root_project_manager(self):
        self.assert_status(200, MANAGER_TOKEN, self.app.get, '/')
        self.assert_status(200, MANAGER_TOKEN, self.app.get, '/v1')

    def test_root_project_service(self):
        self.assert_status(200, SERVICE_TOKEN, self.app.get, '/')
        self.assert_status(200, SERVICE_TOKEN, self.app.get, '/v1')

    def test_root_project_member(self):
        self.assert_status(200, MEMBER_TOKEN, self.app.get, '/')
        self.assert_status(200, MEMBER_TOKEN, self.app.get, '/v1')

    def test_root_project_reader(self):
        self.assert_status(200, READER_TOKEN, self.app.get, '/')
        self.assert_status(200, READER_TOKEN, self.app.get, '/v1')

    def test_root_project_no_role(self):
        self.assert_status(200, NO_ROLE_TOKEN, self.app.get, '/')
        self.assert_status(200, NO_ROLE_TOKEN, self.app.get, '/v1')

    def test_introspect_project_admin(self):
        self.assert_status(202, ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s' % self.uuid)

    def test_introspect_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.post,
                           '/v1/introspection/%s' % self.uuid)

    def test_introspect_project_service(self):
        self.assert_status(202, SERVICE_TOKEN, self.app.post,
                           '/v1/introspection/%s' % self.uuid)

    def test_introspect_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.post,
                           '/v1/introspection/%s' % self.uuid)

    def test_introspect_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_project_admin(self):
        self.assert_status(202, ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_project_service(self):
        self.assert_status(202, SERVICE_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_abort_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.post,
                           '/v1/introspection/%s/abort' % self.uuid)

    def test_status_project_admin(self):
        self.assert_status(200, ADMIN_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_status_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_status_project_service(self):
        self.assert_status(200, SERVICE_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_status_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_status_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.get,
                           '/v1/introspection/%s' % self.uuid)

    def test_list_project_admin(self):
        self.assert_status(200, ADMIN_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_list_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_list_project_service(self):
        self.assert_status(200, SERVICE_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_list_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_list_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.get,
                           '/v1/introspection')

    def test_data_project_admin(self):
        self.assert_status(404, ADMIN_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_project_service(self):
        self.assert_status(404, SERVICE_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.get,
                           '/v1/introspection/%s/data' % self.uuid)

    def test_data_unprocessed_project_admin(self):
        self.assert_status(400, ADMIN_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_data_unprocessed_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_data_unprocessed_project_service(self):
        self.assert_status(400, SERVICE_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_data_unprocessed_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_data_unprocessed_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.post,
                           '/v1/introspection/%s/data/unprocessed' % self.uuid,
                           data={'foo': 'bar'})

    def test_rule_list_project_admin(self):
        self.assert_status(200, ADMIN_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_list_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_list_project_service(self):
        self.assert_status(200, SERVICE_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_list_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_list_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.get,
                           '/v1/rules')

    def test_rule_get_project_admin(self):
        self.assert_status(404, ADMIN_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_get_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_get_project_service(self):
        self.assert_status(404, SERVICE_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_get_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_get_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.get,
                           '/v1/rules/foo')

    def test_rule_delete_all_project_admin(self):
        self.assert_status(204, ADMIN_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_all_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_all_project_service(self):
        self.assert_status(204, SERVICE_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_all_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_all_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.delete,
                           '/v1/rules')

    def test_rule_delete_project_admin(self):
        self.assert_status(404, ADMIN_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_delete_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_delete_project_service(self):
        self.assert_status(404, SERVICE_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_delete_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_delete_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.delete,
                           '/v1/rules/foo')

    def test_rule_create_project_admin(self):
        self.assert_status(500, ADMIN_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })

    def test_rule_create_project_manager(self):
        self.assert_status(403, MANAGER_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })

    def test_rule_create_project_service(self):
        self.assert_status(500, SERVICE_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })

    def test_rule_create_project_member(self):
        self.assert_status(403, MEMBER_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })

    def test_rule_create_project_reader(self):
        self.assert_status(403, READER_TOKEN, self.app.post,
                           '/v1/rules',
                           data={
                               'uuid': self.uuid,
                               'conditions': 'cond',
                               'actions': 'act'
                           })



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""DB models API for inspection data and shared database code."""


import threading
import time

from oslo_config import cfg
from oslo_db import api as oslo_db_api
from oslo_db import exception as db_exc
from oslo_db import options as db_opts
from oslo_db.sqlalchemy import enginefacade
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy import delete
from sqlalchemy import insert
from sqlalchemy import or_, and_
from sqlalchemy import orm
from sqlalchemy.orm import exc as orm_errors
from sqlalchemy import update

from ironic_inspector.common.i18n import _
from ironic_inspector.db import model
from ironic_inspector import utils


LOG = utils.getProcessingLogger(__name__)

_DEFAULT_SQL_CONNECTION = 'sqlite:///ironic_inspector.sqlite'

_CONTEXT = threading.local()

db_opts.set_defaults(cfg.CONF, connection=_DEFAULT_SQL_CONNECTION)

CONF = cfg.CONF


def init():
    """Initialize the database.

    Method called on service start up, initialize transaction
    context manager and try to create db session.
    """
    get_writer_session()


def model_query(model, *args, **kwargs):
    """Query helper for simpler session usage.

    :param session: if present, the session to use
    """
    with session_for_read() as session:
        query = session.query(model, *args)
        return query


def get_writer_session():
    """Help method to get writer session.

    :returns: The writer session.
    """
    return enginefacade.writer.using(_CONTEXT)


def session_for_read():
    """Create read session within context manager"""
    return enginefacade.reader.using(_CONTEXT)


def session_for_write():
    """Create write session within context manager"""
    return enginefacade.writer.using(_CONTEXT)


def get_nodes():
    """Get list of cached nodes

    :returns: list of nodes, could be empty
    """
    with session_for_read() as session:
        res = session.query(
            model.Node
        ).order_by(
                model.Node.started_at.desc()
        )
        return [model.Node(uuid=entry.uuid, version_id=entry.version_id,
                state=entry.state, started_at=entry.started_at,
                finished_at=entry.finished_at, error=entry.error,
                manage_boot=entry.manage_boot)
                for entry in res.all()]


def get_node(uuid, **fields):
    """Get all cached nodes

    :param uuid: node uuid
    :param fields: fields are used as filtering criterion
    :returns: get node object
    :raises: NodeNotFoundInDBError in case node not found or node
             version differ from passed in fields.
    """
    try:
        with session_for_read() as session:
            res = session.query(model.Node).filter_by(
                uuid=uuid, **fields).one()
            return model.Node(uuid=res.uuid, version_id=res.version_id,
                              state=res.state, started_at=res.started_at,
                              finished_at=res.finished_at, error=res.error,
                              manage_boot=res.manage_boot)
    except (orm_errors.NoResultFound, orm_errors.StaleDataError):
        raise utils.NodeNotFoundInDBError()


def get_active_nodes(started_before=None):
    """Get list of nodes on introspection

    :param started_before: datetime object, returns nodes,
                           started before provided time
    :returns: list of nodes, could be empty
    """
    with session_for_read() as session:
        query = session.query(model.Node).filter_by(
            finished_at=None).order_by(model.Node.started_at.desc())

        if started_before:
            query = query.filter(model.Node.started_at < started_before)
        return [model.Node(uuid=entry.uuid, version_id=entry.version_id,
                           state=entry.state, started_at=entry.started_at,
                           finished_at=entry.finished_at, error=entry.error,
                           manage_boot=entry.manage_boot)
                for entry in query.all()]


def list_nodes_by_attributes(attributes):
    """Get list of nodes with certain attributes

    :param attributes: list of attributes as (name, value) pair
    :returns: list of nodes, could be empty
    """
    attr_filters = []
    for name, value in attributes:
        attr_filters.append(and_(model.Attribute.name == name,
                                 model.Attribute.value == value))
    with session_for_read() as session:
        query = session.query(
            model.Attribute
        ).filter(or_(*attr_filters)).all()
        result = [model.Attribute(uuid=attr.uuid, node_uuid=attr.node_uuid,
                                  name=attr.name, value=attr.value)
                  for attr in query]
    return result


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
def update_node(uuid, **values):
    """Update node by uuid

    Updates node fields with provided values, also bump node version.

    :param uuid: node uuid
    :param values: node fields with values to be updated
    :raises: NodeNotFoundInDBError in case node not found or node
             version differ from passed in values.
    """
    fields_ver = values.copy()
    with session_for_write() as session:
        stmt = update(
            model.Node
        ).where(
            model.Node.uuid == uuid
        ).values(
            fields_ver
        ).execution_options(
            synchronize_session=False
        )
        res = session.execute(stmt)
        if res.rowcount == 0:
            raise utils.NodeNotFoundInDBError()


def create_node(uuid, state, started_at=None, finished_at=None,
                error=None, manage_boot=None):
    """Create new node

    :param uuid: node uuid
    :param state: initial node state
    :param started_at: node caching datetime
    :param finished_at: introspection finished datetime
    :param error: introspection error
    :returns: created node object
    """
    node = model.Node(uuid=uuid, state=state, started_at=started_at,
                      finished_at=finished_at,
                      error=error, manage_boot=manage_boot)
    with session_for_write() as session:
        session.add(node)
    return node


def add_node(uuid, state, started_at=None, finished_at=None,
             error=None, manage_boot=None):
    """Add new node

    Before creating new node with certain uuid clean ups all existing
    node info.

    :param uuid: node uuid
    :param state: initial node state
    :param started_at: node caching datetime
    :param finished_at: introspection finished datetime
    :param error: introspection error
    :param manage_boot: whether to manage boot for this node
    :returns: created node object
    """
    with session_for_write() as session:
        # Delete attribute data
        session.execute(
            delete(model.Attribute).where(
                model.Attribute.node_uuid == uuid))
        # Delete introspection data
        session.execute(
            delete(model.Option).where(
                model.Option.uuid == uuid))
        session.execute(
            delete(model.IntrospectionData).where(
                model.IntrospectionData.uuid == uuid))
        # Delete the actual node
        session.execute(
            delete(model.Node).where(
                model.Node.uuid == uuid
            ).execution_options(synchronize_session=False)
        )
        node = model.Node(uuid=uuid, state=state, started_at=started_at,
                          finished_at=finished_at, error=error,
                          manage_boot=manage_boot)
        session.add(node)

    return node


def list_nodes_options_by_uuid(uuid):
    """Get list of node options

    :param uuid: node uuid
    :returns: list of node options, could be empty
    """
    with session_for_read() as session:
        query = session.query(model.Option).filter(model.Option.uuid == uuid)
        return [model.Option(uuid=opt.uuid, name=opt.name, value=opt.value)
                for opt in query.all()]


def delete_node(uuid):
    """Delete node and its attributes

    :param uuid: node uuid
    :returns: None
    """
    with session_for_write() as session:
        # Delete attribute data
        session.execute(
            delete(model.Attribute).where(
                model.Attribute.node_uuid == uuid))
        # Delete introspection data
        session.execute(
            delete(model.Option).where(
                model.Option.uuid == uuid))
        session.execute(
            delete(model.IntrospectionData).where(
                model.IntrospectionData.uuid == uuid))
        # Delete the actual node
        session.execute(
            delete(model.Node).where(
                model.Node.uuid == uuid
            ).execution_options(synchronize_session=False)
        )


def delete_nodes(finished_until=None):
    """Delete all nodes

    :param finished_until: datetime object, delete nodes are
                           introspected before finished_until time
    :returns: None
    """
    with session_for_read() as session:
        query = session.query(model.Node.uuid)
        if finished_until:
            query = query.filter(
                model.Node.finished_at.isnot(None),
                model.Node.finished_at < finished_until)
        uuid_list = []
        for node in query.all():
            # This breaks the requests up and allows proper value
            # deletion since there are structural dependencies on
            # for nodes in other tables. Performance wise this takes
            # a little slower overall, but doesn't cause the tables to
            # be locked, and handles the other tables without building
            # DB triggers.
            uuid_list.append(node[0])
    for uuid in uuid_list:
        delete_node(uuid)
        # Allow the Python GIL to let something else run, and
        # give the DB a chance to breath.
        time.sleep(0)


def set_option(node_uuid, name, value):
    """Set option for node

    :param node_uuid: node uuid
    :param name: option name
    :param value: option value
    :returns: None
    """
    with session_for_write() as session:
        opt = model.Option(uuid=node_uuid, name=name, value=value)
        session.add(opt)


def delete_options(**filters):
    """Delete all options

    :param filters: deletion filter criteria
    :returns: None
    """
    with session_for_write() as session:
        session.query(model.Option).filter_by(**filters).delete()


def set_attribute(node_uuid, name, values):
    """Set lookup attributes for node

    :param node_uuid: node uuid
    :param name: option name
    :param values: list of attribute values
    :returns: None
    """
    if not isinstance(values, list):
        values = [values]
    with session_for_write() as session:

        for value in values:
            attr = model.Attribute(node_uuid=node_uuid,
                                   uuid=uuidutils.generate_uuid(),
                                   name=name, value=value)
            session.add(attr)


def delete_attributes(uuid):
    """Delete all attributes

    :param uuid: the UUID of the node whose attributes you wish
                 tod elete
    :returns: None
    """
    # FIXME(TheJulia): This is going to be difficult to match
    # in later versions of sqlalchemy since query needs to move
    # to use the object model instead of free form attribute name.
    with session_for_write() as session:
        session.execute(
            delete(model.Attribute).where(
                model.Attribute.node_uuid == uuid))


def get_attributes(order_by=None, **fields):
    """Get all attributes

    :param order_by: ordering criterion
    :param fields: filter criteria fields
    :returns: list of attributes
    """
    # FIXME(TheJulia) This needs to be rewritten
    with session_for_read() as session:
        query = session.query(model.Attribute).filter_by(**fields)
        if order_by:
            orders = [getattr(model.Attribute, key) for key in order_by]
            query = query.order_by(*orders)
        res = query.all()

    result = [model.Attribute(uuid=attr.uuid, node_uuid=attr.node_uuid,
                              name=attr.name, value=attr.value)
              for attr in res]
    return result


def get_options(**fields):
    """Get all options

    :param fields: filter criteria fields
    :returns: list of options
    """
    return model_query(model.Option).filter_by(**fields).all()


def create_rule(uuid, conditions, actions, description=None,
                scope=None):
    """Create new rule

    :param uuid: rule uuid
    :param conditions: list of (field, op, multiple, invert, params) tuple,
                       which represents condition object
    :param actions: list of (action, params) pair, which represents action
                    object
    :param description: rule description
    :param scope: rule scope

    :returns: created rule
    """
    try:
        with session_for_write() as session:
            rule = model.Rule(
                uuid=uuid, description=description,
                disabled=False, created_at=timeutils.utcnow(), scope=scope)
            rule.conditions = rule.action = []
            for field, op, multiple, invert, params in conditions:
                rule.conditions.append(model.RuleCondition(op=op,
                                                           field=field,
                                                           multiple=multiple,
                                                           invert=invert,
                                                           params=params))

            for action, params in actions:
                rule.actions.append(model.RuleAction(action=action,
                                                     params=params))

            session.add(rule)
    except db_exc.DBDuplicateEntry as exc:
        LOG.error('Database integrity error %s when creating a rule', exc)
        raise utils.RuleUUIDExistError(uuid)
    return rule


def get_rule(uuid):
    """Get rule by uuid

    :param uuid: rule uuid
    :returns: rule object
    """
    try:
        with session_for_read() as session:
            query = session.query(model.Rule).where(
                model.Rule.uuid == uuid)
            rule = query.one()
            return model.Rule(uuid=rule.uuid, created_at=rule.created_at,
                              description=rule.description,
                              disabled=rule.disabled, scope=rule.scope,
                              conditions=rule.conditions, actions=rule.actions)
    except orm.exc.NoResultFound:
        raise utils.RuleNotFoundError(uuid)


def get_rules(**fields):
    """List all rules."""
    with session_for_read() as session:
        query = session.query(
            model.Rule
        ).filter_by(
            **fields
        ).order_by(
            model.Rule.created_at
        )
        return [model.Rule(
                    uuid=rule.uuid,
                    actions=rule.actions,
                    conditions=rule.conditions,
                    description=rule.description,
                    scope=rule.scope)
                for rule in query]


def get_rules_conditions(**fields):
    """Get all rule conditions

    :param fields: field filter criteria
    :returns: list of conditions
    """
    # NOTE(TheJulia): This appears to exist largely to help unit
    #                 testing of rules functionality.
    with session_for_read() as session:
        query = session.query(
            model.RuleCondition
        ).filter_by(**fields)
        return [model.RuleCondition(
                    id=condition.id,
                    rule=condition.rule,
                    op=condition.op,
                    multiple=condition.multiple,
                    invert=condition.invert,
                    field=condition.field,
                    params=condition.params)
                for condition in query.all()]


def get_rules_actions(**fields):
    """Get all rule actions

    :param fields: field filter criteria
    :returns: list of actions
    """
    # NOTE(TheJulia): This appears to exist largely to help unit
    #                 testing of rules functionality.
    with session_for_read() as session:
        query = session.query(
            model.RuleAction
        ).filter_by(**fields)
        return [model.RuleAction(
                    id=action.id,
                    rule=action.rule,
                    action=action.action,
                    params=action.params)
                for action in query.all()]


def delete_rule(uuid):
    """Delete the rule by uuid

    :param uuid: rule uuid
    :raises: RuleNotFoundError in case rule not found
    :returns: None
    """
    with session_for_write() as session:
        stmt = (
            delete(
                model.RuleAction
            ).where(
               model.RuleAction.rule == uuid
            ).execution_options(synchronize_session=False)
        )
        session.execute(stmt)

        stmt = (
            delete(
                model.RuleCondition
            ).where(
                model.RuleCondition.rule == uuid
            ).execution_options(synchronize_session=False)
        )
        session.execute(stmt)

        stmt = (
            delete(
                model.Rule
            ).where(
                model.Rule.uuid == uuid
            ).execution_options(synchronize_session=False)
        )
        res = session.execute(stmt)
        if res.rowcount == 0:
            raise utils.RuleNotFoundError(uuid)


def delete_all_rules():
    """Delete all rules

    :returns: None
    """
    with session_for_write() as session:
        session.execute(
            delete(model.RuleAction).execution_options(
                synchronize_session=False
            )
        )
        session.execute(
            delete(model.RuleCondition).execution_options(
                synchronize_session=False
            )
        )
        session.execute(
            delete(model.Rule).execution_options(
                synchronize_session=False
            )
        )
        session.commit()


def store_introspection_data(node_id, introspection_data,
                             processed=True):
    """Store introspection data for this node.

    :param node_id: node UUID.
    :param introspection_data: A dictionary of introspection data
    :param processed: Specify the type of introspected data, set to False
                      indicates the data is unprocessed.
    """
    updated = False
    with session_for_write() as session:
        record = session.query(model.IntrospectionData).filter_by(
            uuid=node_id, processed=processed).first()

        if record:
            record.update({'data': introspection_data})
            updated = True
        else:
            # by default, all write sessions are committed. In this
            # case, we can safely rollback. Once we rollback, we
            # launch a new session.
            session.rollback()
    if not updated:
        with session_for_write() as session:
            stmt = insert(model.IntrospectionData).values(
                {'uuid': node_id, 'processed': processed,
                 'data': introspection_data}
            )
            session.execute(stmt)


def get_introspection_data(node_id, processed=True):
    """Get introspection data for this node.

    :param node_id: node UUID.
    :param processed: Specify the type of introspected data, set to False
                      indicates retrieving the unprocessed data.
    :return: A dictionary representation of intropsected data
    """
    try:
        with session_for_read() as session:
            ref = session.query(model.IntrospectionData).filter_by(
                uuid=node_id, processed=processed).one()
        res = ref['data']
        return res
    except orm_errors.NoResultFound:
        msg = _('Introspection data not found for node %(node)s, '
                'processed=%(processed)s') % {'node': node_id,
                                              'processed': processed}
        raise utils.IntrospectionDataNotFound(msg)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base code for PXE boot filtering."""

import contextlib
import functools
import os
import re

from automaton import exceptions as automaton_errors
from automaton import machines
from eventlet import semaphore
from futurist import periodics
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log
import stevedore

from ironic_inspector.common.i18n import _
from ironic_inspector.common import ironic as ir_utils
from ironic_inspector import node_cache
from ironic_inspector.pxe_filter import interface
from ironic_inspector import utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)

_STEVEDORE_DRIVER_NAMESPACE = 'ironic_inspector.pxe_filter'

_EMAC_REGEX = 'EMAC=([0-9a-f]{2}(:[0-9a-f]{2}){5}) IMAC=.*'


class InvalidFilterDriverState(RuntimeError):
    """The fsm of the filter driver raised an error."""


class States(object):
    """PXE filter driver states."""
    uninitialized = 'uninitialized'
    initialized = 'initialized'


class Events(object):
    """PXE filter driver transitions."""
    initialize = 'initialize'
    sync = 'sync'
    reset = 'reset'


# a reset is always possible
State_space = [
    {
        'name': States.uninitialized,
        'next_states': {
            Events.initialize: States.initialized,
            Events.reset: States.uninitialized,
        },
    },
    {
        'name': States.initialized,
        'next_states': {
            Events.sync: States.initialized,
            Events.reset: States.uninitialized,
        },
    },
]


def locked_driver_event(event):
    """Call driver method having processed the fsm event."""
    def outer(method):
        @functools.wraps(method)
        def inner(self, *args, **kwargs):
            with self.lock, self.fsm_reset_on_error() as fsm:
                fsm.process_event(event)
                return method(self, *args, **kwargs)
        return inner
    return outer


class BaseFilter(interface.FilterDriver):
    """The generic PXE boot filtering interface implementation.

    This driver doesn't do anything but provides a basic synchronization and
    initialization logic for some drivers to reuse. Subclasses have to provide
    a custom sync() method.
    """

    fsm = machines.FiniteMachine.build(State_space)
    fsm.default_start_state = States.uninitialized

    def __init__(self):
        super(BaseFilter, self).__init__()
        # Configuration check
        if (CONF.pxe_filter.deny_unknown_macs
                and CONF.processing.node_not_found_hook):
            msg = _('Configuration error: [pxe_filter]deny_unknown_macs is'
                    'enabled and [processing]node_not_found_hook is enabled.'
                    'These options cannot both be enabled simultaneously.')
            raise utils.Error(msg)

        self.lock = semaphore.BoundedSemaphore()
        self.fsm.initialize(start_state=States.uninitialized)

    def __str__(self):
        return '%(driver)s, state=%(state)s' % {
            'driver': type(self).__name__, 'state': self.state}

    @property
    def state(self):
        """Current driver state."""
        return self.fsm.current_state

    def reset(self):
        """Reset internal driver state.

        This method is called by the fsm_context manager upon exception as well
        as by the tear_down_filter method. A subclass might wish to override as
        necessary, though must not lock the driver. The overriding subclass
        should up-call.

        :returns: nothing.
        """
        LOG.debug('Resetting the PXE filter driver %s', self)
        # a reset event is always possible
        self.fsm.process_event(Events.reset)

    @contextlib.contextmanager
    def fsm_reset_on_error(self):
        """Reset the filter driver upon generic exception.

        The context is self.fsm. The automaton.exceptions.NotFound error is
        cast to the InvalidFilterDriverState error. Other exceptions trigger
        self.reset()

        :raises: InvalidFilterDriverState
        :returns: nothing.
        """
        try:
            yield self.fsm
        except automaton_errors.NotFound as e:
            raise InvalidFilterDriverState(_('The PXE filter driver %(driver)s'
                                             ': my fsm encountered an '
                                             'exception: %(error)s') % {
                                                 'driver': self, 'error': e})
        except Exception as e:
            LOG.exception('The PXE filter %(filter)s encountered an '
                          'exception: %(error)s; resetting the filter',
                          {'filter': self, 'error': e})
            self.reset()
            raise

    @locked_driver_event(Events.initialize)
    def init_filter(self):
        """Base driver initialization logic. Locked.

        :raises: InvalidFilterDriverState
        :returns: nothing.
        """
        LOG.debug('Initializing the PXE filter driver %s', self)

    def tear_down_filter(self):
        """Base driver tear down logic. Locked.

        :returns: nothing.
        """
        LOG.debug('Tearing down the PXE filter driver %s', self)
        with self.lock:
            self.reset()

    @locked_driver_event(Events.sync)
    def sync(self, ironic):
        """Base driver sync logic. Locked.

        :param ironic: obligatory ironic client instance
        :returns: nothing.
        """

    def get_periodic_sync_task(self):
        """Get periodic sync task for the filter.

        The periodic task returned is casting the InvalidFilterDriverState
        to the periodics.NeverAgain exception to quit looping.

        :raises: periodics.NeverAgain
        :returns: a periodic task to be run in the background.
        """
        _cached_client = None

        def periodic_sync_task():
            nonlocal _cached_client
            if _cached_client is None:
                _cached_client = ir_utils.get_client()
            try:
                self.sync(_cached_client)
            except InvalidFilterDriverState as e:
                LOG.warning('Filter driver %s disabling periodic sync '
                            'task because of an invalid state.', self)
                raise periodics.NeverAgain(e)

        return periodics.periodic(
            # NOTE(milan): the periodic decorator doesn't support 0 as
            # a spacing value of (a switched off) periodic
            spacing=CONF.pxe_filter.sync_period or float('inf'),
            enabled=bool(CONF.pxe_filter.sync_period))(periodic_sync_task)


class NoopFilter(BaseFilter):
    """A trivial PXE boot filter."""

    def get_periodic_sync_task(self):
        return None


_DRIVER_MANAGER = None


@lockutils.synchronized(__name__)
def _driver_manager():
    """Create a Stevedore driver manager for filtering drivers. Locked."""
    global _DRIVER_MANAGER

    name = CONF.pxe_filter.driver

    if _DRIVER_MANAGER is None:
        _DRIVER_MANAGER = stevedore.driver.DriverManager(
            _STEVEDORE_DRIVER_NAMESPACE,
            name=name,
            invoke_on_load=True
        )

    return _DRIVER_MANAGER


def driver():
    """Get the driver for the PXE filter.

    :returns: the singleton PXE filter driver object.
    """
    return _driver_manager().driver


def _ib_mac_to_rmac_mapping(ports):
    """Update port InfiniBand MAC address to EthernetOverInfiniBand MAC

    On InfiniBand deployment we need to map between the baremetal host
    InfiniBand MAC to the EoIB MAC. The EoIB MAC addresses are learned
    automatically by the EoIB interfaces and those MACs are recorded to the
    /sys/class/net/<ethoib_interface>/eth/neighs file. The InfiniBand GUID is
    taken from the ironic port client-id extra attribute. The InfiniBand GUID
    is the last 8 bytes of the client-id. The file format allows to map the
    GUID to EoIB MAC. The filter rules based on those MACs get applied during a
    driver.update() call

    :param ports: list of ironic ports
    :returns: Nothing.
    """
    ethoib_interfaces = CONF.iptables.ethoib_interfaces
    for iface in ethoib_interfaces:
        neighs_file = (os.path.join('/sys/class/net', iface, 'eth/neighs'))
        try:
            with open(neighs_file, 'r') as fd:
                data = fd.read()
        except IOError:
            LOG.error('Interface %s is not Ethernet Over InfiniBand; '
                      'Skipping ...', iface)
            continue
        for port in ports:
            client_id = port.extra.get('client-id')
            if client_id:
                # Note(moshele): The last 8 bytes in the client-id is
                # the baremetal node InfiniBand GUID
                guid = client_id[-23:]
                p = re.compile(_EMAC_REGEX + guid)
                match = p.search(data)
                if match:
                    port.address = match.group(1)


def get_ironic_macs(ironic):
    ports = [port for port in
             ir_utils.call_with_retries(ironic.ports, limit=None,
                                        fields=['address', 'extra'])]
    _ib_mac_to_rmac_mapping(ports)
    return {port.address for port in ports}


def get_inactive_macs(ironic):
    ports = [port for port in
             ir_utils.call_with_retries(ironic.ports, limit=None,
                                        fields=['address', 'extra'])
             if port.address not in node_cache.active_macs()]
    _ib_mac_to_rmac_mapping(ports)
    return {port.address for port in ports}


def get_active_macs(ironic):
    ports = [port for port in
             ir_utils.call_with_retries(ironic.ports, limit=None,
                                        fields=['address', 'extra'])
             if port.address in node_cache.active_macs()]
    _ib_mac_to_rmac_mapping(ports)
    return {port.address for port in ports}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gather and distinguish Accelerator PCI devices from inventory."""

from oslo_config import cfg
import yaml

from ironic_inspector.plugins import base
from ironic_inspector import utils


CONF = cfg.CONF
LOG = utils.getProcessingLogger(__name__)


class AccelDevicesHook(base.ProcessingHook):
    """Processing hook for distinguishing accelerator devices."""

    def __init__(self):
        super(AccelDevicesHook, self).__init__()
        self._known_devices = {}
        with open(CONF.accelerators.known_devices) as f:
            self._known_devices = yaml.safe_load(f)
        self._validate_datasource()

    def _validate_datasource(self):
        # Do a simple check against the data source
        if (not self._known_devices or
                'pci_devices' not in self._known_devices):
            raise RuntimeError('Could not find pci_devices in the '
                               'configuration data')
        if not isinstance(self._known_devices['pci_devices'], list):
            raise RuntimeError('pci_devices should contain a list of devices')
        for device in self._known_devices['pci_devices']:
            if not device.get('vendor_id') or not device.get('device_id'):
                raise RuntimeError('one of devices is missing vendor_id or '
                                   'device_id')

    def _find_accelerator(self, vendor_id, device_id):
        for dev in self._known_devices['pci_devices']:
            if (dev['vendor_id'] == vendor_id and
                    dev['device_id'] == device_id):
                return dev

    def before_update(self, introspection_data, node_info, **kwargs):
        pci_devices = introspection_data.get('pci_devices', [])
        if not pci_devices:
            LOG.warning('Unable to distinguish accelerator devices due to no '
                        'PCI devices information was received from the '
                        'ramdisk.')
            return

        accelerators = []
        for pci_dev in pci_devices:
            dev = self._find_accelerator(pci_dev['vendor_id'],
                                         pci_dev['product_id'])
            if dev:
                accel = {k: dev[k] for k in dev.keys()}
                accel.update(pci_address=pci_dev['bus'])
                accelerators.append(accel)

        if accelerators:
            node_info.update_properties(accelerators=accelerators)
            LOG.info('Found the following accelerator devices: %s',
                     accelerators)
        else:
            LOG.info('No known accelerator devices found')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Introducing the aborting state

Revision ID: 18440d0834af
Revises: 882b2d84cb1b
Create Date: 2017-12-11 15:40:13.905554

"""

# from alembic import op
# import sqlalchemy as sa
# from sqlalchemy import sql

# from ironic_inspector import introspection_state as istate

# revision identifiers, used by Alembic.
revision = '18440d0834af'
down_revision = '882b2d84cb1b'
branch_labels = None
depends_on = None


# old_state = sa.Enum(*(set(istate.States.all()) - {istate.States.aborting}),
#                     name='node_state')
# new_state = sa.Enum(*istate.States.all(), name='node_state')
# Node = sql.table('nodes', sql.column('state', old_state))


def upgrade():
    # NOTE(TheJulia): Batch table alteration generally works by making
    # a copy of the table, modifying the copy, and dropping the original,
    # and renaming the updated copy to be the same name. Unfortunately
    # this results in a "FOREIGN KEY constraint failed" error in
    # newer versions of inspector as two additional tables now exist
    # which did not exist when the upgrade was written.
    #
    # with op.batch_alter_table('nodes') as batch_op:
    #     batch_op.alter_column('state', existing_type=old_state,
    #                           type_=new_state)
    #
    # The prior net effect was that the field was being altered to
    # be varchar(10), to house the largest enum value, except the added
    # value only requires varchar(9), so this is sort of entirely
    # redundant at this point. For what it is worth, while
    # batch_alter_table *does* include an option to prevent
    # recration, column widths cannot be modified dynamically with
    # sqlite, which results in a different exception if attempted.
    pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from ironic_inspector.common import lldp_parsers as nv
from ironic_inspector.plugins import lldp_basic
from ironic_inspector.test import base as test_base


class TestLLDPBasicProcessingHook(test_base.NodeTest):
    hook = lldp_basic.LLDPBasicProcessingHook()

    def setUp(self):
        super(TestLLDPBasicProcessingHook, self).setUp()
        self.data = {
            'inventory': {
                'interfaces': [{
                    'name': 'em1',
                }],
                'cpu': 1,
                'disks': 1,
                'memory': 1
            },
            'all_interfaces':
            {
                'em1': {'mac': self.macs[0], 'ip': self.ips[0]}
            }
        }

        self.expected = {"em1": {"ip": self.ips[0], "mac": self.macs[0]}}

    def test_all_valid_data(self):
        self.data['lldp_raw'] = {
            'em1': [
                [1, "04112233aabbcc"],  # ChassisId
                [2, "07373334"],        # PortId
                [3, "003c"],            # TTL
                [4, "686f737430322e6c61622e656e6720706f7274203320"
                 "28426f6e6429"],  # PortDesc
                [5, "737730312d646973742d31622d623132"],  # SysName
                [6, "4e6574776f726b732c20496e632e20353530302c2076657273696f"
                 "6e203132204275696c6420646174653a20323031342d30332d31332030"
                 "383a33383a33302055544320"],  # SysDesc
                [7, "00140014"],  # SysCapabilities
                [8, "0501c000020f020000000000"],  # MgmtAddress
                [8, "110220010db885a3000000008a2e03707334020000000000"],
                [8, "0706aa11bb22cc3302000003e900"],  # MgmtAddress
                [127, "00120f01036c110010"],  # dot3 MacPhyConfigStatus
                [127, "00120f030300000002"],  # dot3 LinkAggregation
                [127, "00120f0405ea"],  # dot3 MTU
                [127, "0080c2010066"],  # dot1 PortVlan
                [127, "0080c20206000a"],  # dot1 PortProtocolVlanId
                [127, "0080c202060014"],  # dot1 PortProtocolVlanId
                [127, "0080c204080026424203000000"],   # dot1 ProtocolIdentity
                [127, "0080c203006507766c616e313031"],  # dot1 VlanName
                [127, "0080c203006607766c616e313032"],  # dot1 VlanName
                [127, "0080c203006807766c616e313034"],  # dot1 VlanName
                [127, "0080c2060058"],  # dot1 MgmtVID
                [0, ""],
            ]
        }
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [[0, ""]]
        }]

        expected = {
            nv.LLDP_CAP_ENABLED_NM: ['Bridge', 'Router'],
            nv.LLDP_CAP_SUPPORT_NM: ['Bridge', 'Router'],
            nv.LLDP_CHASSIS_ID_NM: "11:22:33:aa:bb:cc",
            nv.LLDP_MGMT_ADDRESSES_NM: ['192.0.2.15',
                                        '2001:db8:85a3::8a2e:370:7334',
                                        'aa:11:bb:22:cc:33'],
            nv.LLDP_PORT_LINK_AUTONEG_ENABLED_NM: True,
            nv.LLDP_PORT_DESC_NM: 'host02.lab.eng port 3 (Bond)',
            nv.LLDP_PORT_ID_NM: '734',
            nv.LLDP_PORT_LINK_AGG_ENABLED_NM: True,
            nv.LLDP_PORT_LINK_AGG_ID_NM: 2,
            nv.LLDP_PORT_LINK_AGG_SUPPORT_NM: True,
            nv.LLDP_PORT_MGMT_VLANID_NM: 88,
            nv.LLDP_PORT_MAU_TYPE_NM: '100BASE-TX full duplex',
            nv.LLDP_MTU_NM: 1514,
            nv.LLDP_PORT_CAPABILITIES_NM: ['1000BASE-T fdx',
                                           '100BASE-TX fdx',
                                           '100BASE-TX hdx',
                                           '10BASE-T fdx',
                                           '10BASE-T hdx',
                                           'Asym and Sym PAUSE fdx'],
            nv.LLDP_PORT_PROT_VLAN_ENABLED_NM: True,
            nv.LLDP_PORT_PROT_VLANIDS_NM: [10, 20],
            nv.LLDP_PORT_PROT_VLAN_SUPPORT_NM: True,
            nv.LLDP_PORT_VLANID_NM: 102,
            nv.LLDP_PORT_VLANS_NM: [{'id': 101, 'name': 'vlan101'},
                                    {'id': 102, 'name': 'vlan102'},
                                    {'id': 104, "name": 'vlan104'}],
            nv.LLDP_PROTOCOL_IDENTITIES_NM: ['0026424203000000'],
            nv.LLDP_SYS_DESC_NM: 'Networks, Inc. 5500, version 12'
            ' Build date: 2014-03-13 08:38:30 UTC ',
            nv.LLDP_SYS_NAME_NM: 'sw01-dist-1b-b12'
        }

        self.hook.before_update(self.data, self.node_info)

        actual_all_int = self.data['all_interfaces']
        actual = actual_all_int['em1']['lldp_processed']

        for name, value in expected.items():
            if name is nv.LLDP_PORT_VLANS_NM:
                for d1, d2 in zip(expected[name], actual[name]):
                    for key, value in d1.items():
                        self.assertEqual(d2[key], value)
            else:
                self.assertEqual(actual[name], expected[name])

    def test_old_format(self):

        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [1, "04112233aabbcc"],  # ChassisId
                [2, "07373334"],        # PortId
                [3, "003c"],            # TTL
                [4, "686f737430322e6c61622e656e6720706f7274203320"
                 "28426f6e6429"],  # PortDesc
                [5, "737730312d646973742d31622d623132"],  # SysName
                [6, "4e6574776f726b732c20496e632e20353530302c2076657273696f"
                 "6e203132204275696c6420646174653a20323031342d30332d31332030"
                 "383a33383a33302055544320"],  # SysDesc
                [7, "00140014"],  # SysCapabilities
                [8, "0501c000020f020000000000"],  # MgmtAddress
                [8, "110220010db885a3000000008a2e03707334020000000000"],
                [8, "0706aa11bb22cc3302000003e900"],  # MgmtAddress
                [127, "00120f01036c110010"],  # dot3 MacPhyConfigStatus
                [127, "00120f030300000002"],  # dot3 LinkAggregation
                [127, "00120f0405ea"],  # dot3 MTU
                [127, "0080c2010066"],  # dot1 PortVlan
                [127, "0080c20206000a"],  # dot1 PortProtocolVlanId
                [127, "0080c202060014"],  # dot1 PortProtocolVlanId
                [127, "0080c204080026424203000000"],   # dot1 ProtocolIdentity
                [127, "0080c203006507766c616e313031"],  # dot1 VlanName
                [127, "0080c203006607766c616e313032"],  # dot1 VlanName
                [127, "0080c203006807766c616e313034"],  # dot1 VlanName
                [127, "0080c2060058"],  # dot1 MgmtVID
                [0, ""]]
            }]

        expected = {
            nv.LLDP_CAP_ENABLED_NM: ['Bridge', 'Router'],
            nv.LLDP_CAP_SUPPORT_NM: ['Bridge', 'Router'],
            nv.LLDP_CHASSIS_ID_NM: "11:22:33:aa:bb:cc",
            nv.LLDP_MGMT_ADDRESSES_NM: ['192.0.2.15',
                                        '2001:db8:85a3::8a2e:370:7334',
                                        'aa:11:bb:22:cc:33'],
            nv.LLDP_PORT_LINK_AUTONEG_ENABLED_NM: True,
            nv.LLDP_PORT_DESC_NM: 'host02.lab.eng port 3 (Bond)',
            nv.LLDP_PORT_ID_NM: '734',
            nv.LLDP_PORT_LINK_AGG_ENABLED_NM: True,
            nv.LLDP_PORT_LINK_AGG_ID_NM: 2,
            nv.LLDP_PORT_LINK_AGG_SUPPORT_NM: True,
            nv.LLDP_PORT_MGMT_VLANID_NM: 88,
            nv.LLDP_PORT_MAU_TYPE_NM: '100BASE-TX full duplex',
            nv.LLDP_MTU_NM: 1514,
            nv.LLDP_PORT_CAPABILITIES_NM: ['1000BASE-T fdx',
                                           '100BASE-TX fdx',
                                           '100BASE-TX hdx',
                                           '10BASE-T fdx',
                                           '10BASE-T hdx',
                                           'Asym and Sym PAUSE fdx'],
            nv.LLDP_PORT_PROT_VLAN_ENABLED_NM: True,
            nv.LLDP_PORT_PROT_VLANIDS_NM: [10, 20],
            nv.LLDP_PORT_PROT_VLAN_SUPPORT_NM: True,
            nv.LLDP_PORT_VLANID_NM: 102,
            nv.LLDP_PORT_VLANS_NM: [{'id': 101, 'name': 'vlan101'},
                                    {'id': 102, 'name': 'vlan102'},
                                    {'id': 104, "name": 'vlan104'}],
            nv.LLDP_PROTOCOL_IDENTITIES_NM: ['0026424203000000'],
            nv.LLDP_SYS_DESC_NM: 'Networks, Inc. 5500, version 12'
            ' Build date: 2014-03-13 08:38:30 UTC ',
            nv.LLDP_SYS_NAME_NM: 'sw01-dist-1b-b12'
        }

        self.hook.before_update(self.data, self.node_info)

        actual_all_int = self.data['all_interfaces']
        actual = actual_all_int['em1']['lldp_processed']

        for name, value in expected.items():
            if name is nv.LLDP_PORT_VLANS_NM:
                for d1, d2 in zip(expected[name], actual[name]):
                    for key, value in d1.items():
                        self.assertEqual(d2[key], value)
            else:
                self.assertEqual(actual[name], expected[name])

    def test_multiple_interfaces(self):
        self.data = {
            # An artificial mix of old and new LLDP fields.
            'inventory': {
                'interfaces': [
                    {'name': 'em1'},
                    {'name': 'em2',
                     'lldp': [
                         [1, "04112233aabbdd"],
                         [2, "07373838"],
                         [3, "003c"]]},
                    {'name': 'em3',
                     'lldp': [
                         [3, "003c"]]}],
                'cpu': 1,
                'disks': 1,
                'memory': 1
            },
            'all_interfaces': {
                'em1': {'mac': self.macs[0], 'ip': self.ips[0]},
                'em2': {'mac': self.macs[0], 'ip': self.ips[0]},
                'em3': {'mac': self.macs[0], 'ip': self.ips[0]}
            },
            'lldp_raw': {
                'em1': [
                    [1, "04112233aabbcc"],
                    [2, "07373334"],
                    [3, "003c"]
                ],
                'em3': [
                    [1, "04112233aabbee"],
                    [2, "07373939"],
                    [3, "003c"]
                ],
            },
        }

        expected = {"em1": {"ip": self.ips[0], "mac": self.macs[0],
                            "lldp_processed": {
                                nv.LLDP_CHASSIS_ID_NM: "11:22:33:aa:bb:cc",
                                nv.LLDP_PORT_ID_NM: "734"}},
                    "em2": {"ip": self.ips[0], "mac": self.macs[0],
                            "lldp_processed": {
                                nv.LLDP_CHASSIS_ID_NM: "11:22:33:aa:bb:dd",
                                nv.LLDP_PORT_ID_NM: "788"}},
                    "em3": {"ip": self.ips[0], "mac": self.macs[0],
                            "lldp_processed": {
                                nv.LLDP_CHASSIS_ID_NM: "11:22:33:aa:bb:ee",
                                nv.LLDP_PORT_ID_NM: "799"}}}

        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(expected, self.data['all_interfaces'])

    def test_chassis_ids(self):
        # Test IPv4 address
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [1, "0501c000020f"],
                ]}]

        self.expected['em1']['lldp_processed'] = {
            nv.LLDP_CHASSIS_ID_NM: "192.0.2.15"
            }
        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])

        # Test name
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [1, "0773773031"],
            ]}]

        self.expected['em1']['lldp_processed'] = {
            nv.LLDP_CHASSIS_ID_NM: "sw01"
        }
        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])

    def test_duplicate_tlvs(self):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [1, "04112233aabbcc"],  # ChassisId
                [1, "04332211ddeeff"],  # ChassisId
                [1, "04556677aabbcc"],  # ChassisId
                [2, "07373334"],  # PortId
                [2, "07373435"],  # PortId
                [2, "07373536"]   # PortId
                ]}]

        # Only the first unique TLV is processed
        self.expected['em1']['lldp_processed'] = {
            nv.LLDP_CHASSIS_ID_NM: "11:22:33:aa:bb:cc",
            nv.LLDP_PORT_ID_NM: "734"
            }

        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])

    def test_unhandled_tlvs(self):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [10, "04112233aabbcc"],
                [12, "07373334"],
                [128, "00120f080300010000"]]}]

        # nothing should be written to lldp_processed
        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])

    def test_unhandled_oui(self):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [127, "00906901425030323134323530393236"],
                [127, "23ac0074657374"],
                [127, "00120e010300010000"]]}]

        # nothing should be written to lldp_processed
        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])

    @mock.patch.object(nv.LOG, 'warning', autospec=True)
    def test_null_strings(self, mock_log):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [1, "04"],
                [4, ""],  # PortDesc
                [5, ""],  # SysName
                [6, ""],  # SysDesc
                [127, "0080c203006507"]  # dot1 VlanName
            ]}]

        self.expected['em1']['lldp_processed'] = {
            nv.LLDP_PORT_DESC_NM: '',
            nv.LLDP_SYS_DESC_NM: '',
            nv.LLDP_SYS_NAME_NM: ''
        }

        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])
        self.assertEqual(2, mock_log.call_count)

    @mock.patch.object(nv.LOG, 'warning', autospec=True)
    def test_truncated_int(self, mock_log):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [127, "00120f04"],  # dot3 MTU
                [127, "0080c201"],  # dot1 PortVlan
                [127, "0080c206"],  # dot1 MgmtVID
                ]}]

        # nothing should be written to lldp_processed
        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])
        self.assertEqual(3, mock_log.call_count)

    @mock.patch.object(nv.LOG, 'warning', autospec=True)
    def test_invalid_ip(self, mock_log):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [8, "0501"],  # truncated
                [8, "0507c000020f020000000000"]]  # invalid id
        }]
        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])
        self.assertEqual(1, mock_log.call_count)

    @mock.patch.object(nv.LOG, 'warning', autospec=True)
    def test_truncated_mac(self, mock_log):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [8, "0506"]]
        }]

        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])
        self.assertEqual(1, mock_log.call_count)

    @mock.patch.object(nv.LOG, 'warning', autospec=True)
    def test_bad_value_macphy(self, mock_log):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [127, "00120f01036c11FFFF"],  # invalid mau type
                [127, "00120f01036c11"],      # truncated
                [127, "00120f01036c"]         # truncated
            ]}]

        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])
        self.assertEqual(3, mock_log.call_count)

    @mock.patch.object(nv.LOG, 'warning', autospec=True)
    def test_bad_value_linkagg(self, mock_log):
        self.data['inventory']['interfaces'] = [{
            'name': 'em1',
            'lldp': [
                [127, "00120f0303"],  # dot3 LinkAggregation
                [127, "00120f03"]     # truncated
                ]}]

        self.hook.before_update(self.data, self.node_info)
        self.assertEqual(self.expected, self.data['all_interfaces'])
        self.assertEqual(2, mock_log.call_count)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gather and distinguish PCI devices from inventory."""

import collections
import json

from oslo_config import cfg

from ironic_inspector.plugins import base
from ironic_inspector import utils


CONF = cfg.CONF

LOG = utils.getProcessingLogger(__name__)


def _parse_pci_alias_entry():
    parsed_pci_devices = []
    for pci_alias_entry in CONF.pci_devices.alias:
        try:
            parsed_entry = json.loads(pci_alias_entry)
            if set(parsed_entry) != {'vendor_id', 'product_id', 'name'}:
                raise KeyError("The 'alias' entry should contain "
                               "exactly 'vendor_id', 'product_id' and "
                               "'name' keys")
            parsed_pci_devices.append(parsed_entry)
        except (ValueError, KeyError) as ex:
            LOG.error("Error parsing 'alias' option: %s", ex)
    return {(dev['vendor_id'], dev['product_id']): dev['name']
            for dev in parsed_pci_devices}


class PciDevicesHook(base.ProcessingHook):
    """Processing hook for counting and distinguishing various PCI devices.

        That information can be later used by nova for node scheduling.
    """
    aliases = _parse_pci_alias_entry()

    def _found_pci_devices_count(self, found_pci_devices):
        return collections.Counter([(dev['vendor_id'], dev['product_id'])
                                    for dev in found_pci_devices
                                    if (dev['vendor_id'], dev['product_id'])
                                    in self.aliases])

    def before_update(self, introspection_data, node_info, **kwargs):
        if 'pci_devices' not in introspection_data:
            if CONF.pci_devices.alias:
                LOG.warning('No PCI devices information was received from '
                            'the ramdisk.')
            return
        alias_count = {self.aliases[id_pair]: count for id_pair, count in
                       self._found_pci_devices_count(
                           introspection_data['pci_devices']).items()}
        if alias_count:
            node_info.update_capabilities(**alias_count)
            LOG.info('Found the following PCI devices: %s', alias_count)
